from ..requests import RequestsClient
from typing import Optional
from .models import AccountInfo, OperationHistory, OperationDetails, RequestPayment, ProcessPayment
from urllib.parse import urlencode


class AsyncYoomoney(RequestsClient):
    API_HOST: str = "https://yoomoney.ru"

    def __init__(self, access_token: str) -> None:
        """
        Initialize Yoomoney API client
        :param apiKey: Your API key
        """
        super().__init__()
        self.__access_token = access_token
        self.__headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "Authorization": f"Bearer {self.__access_token}",
        }
        self.__base_url = "https://yoomoney.ru/api"
        self.__post_method = "POST"
        self.__payment_name = "yoomoney"
        self.__payment_name_quick_pay = "yoomoney_quick-pay"
        self.check_values()

    def check_values(self):
        if not self.__access_token:
            raise ValueError('No access token specified')
        
    async def account_info(self) -> AccountInfo:
        """Obtaining information about the user's account status.
        
        Docs: https://yoomoney.ru/docs/wallet/user-account/account-info"""
        url = f"{self.__base_url}/account-info"
        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers)
        return AccountInfo(**response)
    
    async def operation_history(self,
                                operation_type: Optional[str] = None,
                                label: Optional[str] = None,
                                date_from: Optional[str] = None,
                                date_till: Optional[str] = None,
                                start_record: Optional[str] = None,
                                records: Optional[int] = None,
                                details: Optional[bool] = None) -> OperationHistory:
        """This method allows you to view transaction history (in whole or in part) page by page. History entries are displayed in reverse chronological order: from most recent to earliest.
        
        :param operation_type: Optional. A list of transaction types to display. Possible values: deposition — account replenishment (income); payment — payments from the account (expense). Transaction types are listed separated by spaces. If this parameter is omitted, all transactions are displayed.
        :param label: Optional. Filtering payments by label value. Payments that have the specified value for the label parameter of the request-payment call are selected.
        :param date_from: Optional. Print transactions from a given time (transactions equal to or later than from ). If the parameter is omitted, all transactions are printed.
        :param date_till: Optional. Display operations up to the time point (operations earlier than 'till'). If the parameter is omitted, all operations are displayed.
        :param start_record: Optional. If this parameter is present, operations will be displayed starting with the start_record number. Operations are numbered starting from 0.
        :param records: Optional. The number of transaction history records to retrieve. Valid values: 1 to 100, default: 30.
        :param details: Optional. Show detailed operation details. Defaults to false. Displaying operation details requires the operation-details permission.
        
        Docs: https://yoomoney.ru/docs/wallet/user-account/operation-history"""
        url = f"{self.__base_url}/operation-history"
        params = {
            "type": operation_type,
            "label": label,
            "from": date_from,
            "till": date_till,
            "start_record": start_record,
            "records": records,
            "details": details
        }
        self._delete_empty_fields(params)
        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=urlencode(params))
        return OperationHistory(**response)
    
    async def operation_details(self, operation_id: Optional[str] = None) -> OperationDetails:
        """Allows you to obtain detailed information about an operation from the history.
        
        :param operation_id: Operation ID. The parameter value should be specified as the value of the operation_id parameter in the operation-history method response or the value of the payment_id field in the process-payment method response if the payer's account history is being requested.

        Docs: https://yoomoney.ru/docs/wallet/user-account/operation-details"""
        url = f"{self.__base_url}/operation-details"
        params = {
            "operation_id": operation_id
        }
        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=urlencode(params))
        return OperationDetails(**response)
    
    async def request_payment(self, 
                              to: str,
                              amount: float,
                              amount_due: Optional[float] = None,
                              comment: Optional[str] = None,
                              message: Optional[str] = None,
                              label: Optional[str] = None,
                              ) -> RequestPayment:
        """Create a payment, check the parameters, and verify that the store can accept the payment or transfer funds to the user's YooMoney account.
        
        :param to: Recipient ID (account number, phone number or email).
        :param amount: Amount to be paid (how much the sender will pay).
        :param amount_due: Optional. Amount to be received (will be credited to the recipient's account after payment).
        :param comment: Optional. A comment on the translation is displayed in the sender's history.
        :param message: Optional. A comment on the transfer, displayed to the recipient.
        :param label: Optional. Payment label.
        
        Docs: https://yoomoney.ru/docs/wallet/process-payments/request-payment"""
        url = f"{self.__base_url}/request-payment"
        params = {
            "pattern_id": "p2p",
            "to": to,
            "amount": amount,
            "amount_due": amount_due,
            "comment": comment,
            "message": message,
            "label": label,
        }
        self._delete_empty_fields(params)
        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=urlencode(params))
        return RequestPayment(**response)
        
    async def process_payment(self, 
                              request_id: str,
                              money_source: Optional[str] = "wallet",
                              csc: Optional[str] = None,
                              ext_auth_success_uri: Optional[str] = None,
                              ext_auth_fail_uri: Optional[str] = None,
                              ) -> ProcessPayment:
        """Confirmation of a payment previously created using the request_payment() method. Specifying the payment method.
        
        :param request_id: The request ID obtained from the response of the request_payment() method.
        :param money_source: Optional. Requested payment method: wallet — from the user's account. ID of the card linked to the account (the value of the id field in the bank card description). Default: wallet.
        :param csc: Optional. Card Security Code (CVV2/CVC2)—the code of the user's linked bank card. This parameter should only be specified when paying with a linked bank card.
        :param ext_auth_success_uri: Optional. The return page address upon successful 3-D Secure payment authentication. Specified if the app supports 3-D Secure authentication. This parameter is required for this type of authentication.
        :param ext_auth_fail_uri: Optional. The return page address if 3-D Secure payment authentication by bank card is denied. This is specified if the app supports 3-D Secure authentication. This parameter is required for this type of authentication.
        
        Docs: https://yoomoney.ru/docs/wallet/process-payments/process-payment"""
        url = f"{self.__base_url}/process-payment"
        params = {
            "request_id": request_id,
            "money_source": money_source,
            "csc": csc,
            "ext_auth_success_uri": ext_auth_success_uri,
            "ext_auth_fail_uri": ext_auth_fail_uri,
        }
        self._delete_empty_fields(params)
        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=urlencode(params))
        return ProcessPayment(**response)
    
    async def quick_pay(self,
                        receiver: str,
                        sum: float,
                        quickpay_form: Optional[str] = "shop",
                        payment_type: Optional[str] = "SB",
                        label: Optional[str] = None,
                        success_url: Optional[str] = None,
                        ) -> str:
        """A form is a set of fields with information about a transfer. You can place the form in your interface (for example, on a website or blog). When the sender clicks the button, the form data is sent to YuMoney and initiates a transfer instruction to your wallet.
        
        :param receiver: Number of the YooMoney wallet which money from senders is credited to.
        :param sum: Transfer amount (the amount debited from the sender).
        :param quickpay_form: Optional. Form type. Fixed value: button/shop.
        :param payment_type: Optional. Payment method. Possible values: PC for a payment from a YooMoney wallet; AC for a payment from a bank card; SB.
        :param label: Optional. The label that a site or app assigns to a certain transfer. For instance, a code or order identifier may be used for this label.
        :param success_url: Optional. URL where the user is redirected after the transfer.
        
        Docs: https://yoomoney.ru/docs/payment-buttons/using-api/forms"""
        url = "https://yoomoney.ru/quickpay/confirm"
        params = {
            "receiver": receiver,
            "sum": sum,
            "quickpay-form": quickpay_form,
            "paymentType": payment_type,
            "label": label,
            "successURL": success_url,
        }
        self._delete_empty_fields(params)
        response = await self._request(self.__payment_name_quick_pay, self.__post_method, url, headers=self.__headers, data=urlencode(params))
        return response
    
    async def check_yoomoney_payment(self, label: str) -> bool:
        """
        Checking payment in the transaction history by label.

        :param label: Label of operation."""
        operations = await self.operation_history(label=label, records=1)
        if operations.operations:
            return operations.operations[0].label == label
        else:
            return False


from pydantic import BaseModel
from typing import Optional


class OrderStatuses:
    PENDING: str = "pending"
    APPROVE: str = "approve"
    DECLINE: str = "decline"
    EXPIRED: str = "expired"


class Order(BaseModel):
    status: Optional[bool] = None
    url: Optional[str] = None
    
    
class OrderInfo(BaseModel):
    status: Optional[bool] = None
    order_status: Optional[str] = None


from ..requests import RequestsClient
from typing import Optional, List
from .models import Balance, Balances, CreatePayment, GenerateStaticWallet, GenerateQrCode, BlockStaticWallet, RefundPaymentsOnBlockedAddress, \
                    PaymentInfo, ListOfServices, PaymentHistory, Payout, PayoutHistory, ListOfServicesPayout, TransferWallet, RecurringPayment, \
                    ListOfRecurringPayments, ExchangeRatesList, Discount
import base64
import json
import hashlib


class AsyncCryptomus(RequestsClient):
    API_HOST: str = "https://cryptomus.com/gateway"

    def __init__(self, payment_api_key: str, merchant_id: str, payout_api_key: str) -> None:
        """
        Initialize Cryptomus API client
        :param payment_api_key: Your payment API key
        :param merchant_id: Your merchant ID
        :param payout_api_key: Your payout API key
        """
        super().__init__()
        self.__payment_api_key = payment_api_key
        self.__merchant_id = merchant_id
        self.__payout_api_key = payout_api_key
        self.__headers = {
            "Content-Type": "application/json",
            "merchant": self.__merchant_id,
        }
        self.__base_url = "https://api.cryptomus.com/v1"
        self.__post_method = "POST"
        self.__get_method = "GET"
        self.__payment_name = "cryptomus"
        self.check_values()

    def check_values(self):
        if not self.__merchant_id or not self.__payment_api_key or not self.__payout_api_key:
            raise ValueError('No Payment API key, merchant ID or Payout API key specified')
        
    def __generate_sign(self, data: Optional[dict] = None, is_for_payouts: Optional[bool] = False) -> dict:
        if data:
            data_encoded = base64.b64encode(json.dumps(data).encode()).decode()
        else:
            data_encoded = ""

        if is_for_payouts:
            return hashlib.md5((data_encoded + self.__payout_api_key).encode()).hexdigest()
        return hashlib.md5((data_encoded + self.__payment_api_key).encode()).hexdigest()

    async def get_balance(self) -> Balances:
        """Get list of your balances.

        Docs: https://doc.cryptomus.com/ru/business/balance"""

        self.__headers["sign"] = self.__generate_sign()
        response = await self._request(self.__payment_name, self.__post_method, f'{self.__base_url}/balance', headers=self.__headers)        
        
        return Balances(merchant=[Balance(**balance) for balance in response['result'][0]['balance']['merchant']], 
                        user=[Balance(**balance) for balance in response['result'][0]['balance']['user']])

    async def create_payment(self, amount: str, currency: str, order_id: str, network: Optional[str] = None, url_return: Optional[str] = None,
                             url_success: Optional[str] = None, url_callback: Optional[str] = None, is_payment_multiple: Optional[bool] = True,
                             lifetime: Optional[int] = 3600, to_currency: Optional[str] = None, subtract: Optional[int] = 0,
                             accuracy_payment_percent: Optional[int] = 0, additional_data: Optional[str] = None, currencies: Optional[list] = None,
                             except_currencies: Optional[list] = None, course_source: Optional[str] = None, from_referral_code: Optional[str] = None,
                             discount_percent: Optional[int] = None, is_refresh: Optional[bool] = False) -> CreatePayment:
        
        """Create payment.
        
        :param amount: Amount to be paid. If there are pennies in the amount, then send them with a separator '.'. Example: 10.28
        :param currency: Currency code.
        :param order_id: Order ID in your system. The parameter should be a string consisting of alphabetic characters, numbers, underscores, and dashes. It should not contain any spaces or special characters. The order_id must be unique within the merchant invoices/static wallets/recurrence payments. When we find an existing invoice with order_id, we return its details, a new invoice will not be created.
        :param network: Blockchain network code.
        :param url_return: Before paying, the user can click on the button on the payment form and return to the store page at this URL.
        :param url_success: After successful payment, the user can click on the button on the payment form and return to this URL.
        :param url_callback: Url to which webhooks with payment status will be sent.
        :param is_payment_multiple: Whether the user is allowed to pay the remaining amount. This is useful when the user has not paid the entire amount of the invoice for one transaction, and you want to allow him to pay up to the full amount. If you disable this feature, the invoice will finalize after receiving the first payment and you will receive funds to your balance.
        :param lifetime: Min: 300. Max: 43200. The lifespan of the issued invoice (in seconds).
        :param to_currency: The parameter is used to specify the target currency for converting the invoice amount. When creating an invoice, you provide an amount and currency, and the API will convert that amount to the equivalent value in the to_currency. For example, to create an invoice for 20 USD in bitcoin: amount: 20, currency: USD, to_currency: BTC. The API will convert 20 USD amount to its equivalent in BTC based on the current exchange rate and the user will pay in BTC. The to_currency should always be the cryptocurrency code, not a fiat currency code.
        :param subtract: Min: 0. Max: 100. Percentage of the payment commission charged to the client. If you have a rate of 1%, then if you create an invoice for 100 USDT with subtract = 100 (the client pays 100% commission), the client will have to pay 101 USDT.
        :param accuracy_payment_percent: Min: 0. Max: 5. Acceptable inaccuracy in payment. For example, if you pass the value 5, the invoice will be marked as Paid even if the client has paid only 95% of the amount. The actual payment amount will be credited to the balance.
        :param additional_data: Max: 255. Additional information for you (not shown to the client).
        :param currencies: List of allowed currencies for payment. This is useful if you want to limit the list of coins that your customers can use to pay invoices.
        :param except_currencies: List of excluded currencies for payment.
        :param course_source: Min: 4. Max: 20. The service from which the exchange rates are taken for conversion in the invoice.
        :param from_referral_code: The merchant who makes the request connects to a referrer by code. For example, you are an application that generates invoices via the Cryptomus API and your customers are other stores. They enter their api key and merchant id in your application, and you send requests with their credentials and passing your referral code. Thus, your clients become referrals on your Cryptomus account and you will receive income from their turnover.
        :param discount_percent: Min: -99. Max: 100. Positive numbers: Allows you to set a discount. To set a 5% discount for the payment, you should pass a value: 5. Negative numbers: Allows you to set custom additional commission. To set an additional commission of 10% for the payment, you should pass a value: -10. The discount percentage when creating an invoice is taken into account only if the invoice has a specific cryptocurrency.
        :param is_refresh: Using this parameter, you can update the lifetime and get a new address for the invoice if the lifetime has expired. To do that, you need to pass all required parameters, and the invoice with passed order_id will be refreshed. Only address, payment_status and expired_at are changed. No other fields are changed, regardless of the parameters passed.
        
        Docs: https://doc.cryptomus.com/business/payments/creating-invoice
        """

        url = f"{self.__base_url}/payment"
        params = {
            "amount": amount,
            "currency": currency,
            "order_id": order_id,
            "network": network,
            "url_return": url_return,
            "url_success": url_success,
            "url_callback": url_callback,
            "is_payment_multiple": is_payment_multiple,
            "lifetime": lifetime,
            "to_currency": to_currency,
            "subtract": subtract,
            "accuracy_payment_percent": accuracy_payment_percent,
            "additional_data": additional_data,
            "currencies": currencies,
            "except_currencies": except_currencies,
            "course_source": course_source,
            "from_referral_code": from_referral_code,
            "discount_percent": discount_percent,
            "is_refresh": is_refresh,
        }
        self._delete_empty_fields(params)
        self.__headers["sign"] = self.__generate_sign(params)

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, json=params)

        return CreatePayment(**response['result'])
    
    async def payment_info(self, uuid: Optional[str] = None, order_id: Optional[str] = None) -> PaymentInfo:
        """Payment information.

        :param uuid: Invoice uuid.
        :param order_id: Invoice order ID.

        Docs: https://doc.cryptomus.com/business/payments/payment-information
        """
        url = f"{self.__base_url}/payment/info"
        params = {
            "uuid": uuid,
            "order_id": order_id,
        }
        self._delete_empty_fields(params)
        self.__headers["sign"] = self.__generate_sign(params)

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, json=params)

        return PaymentInfo(**response['result'])

    async def generate_static_wallet(self, currency: str, network: str, order_id: str, url_callback: Optional[str] = None, 
                                     from_referral_code: Optional[str] = None) -> GenerateStaticWallet:
        """Creating a Static wallet.
        
        :param currency: Currency code.
        :param network: Blockchain network code.
        :param order_id: Order ID in your system. The parameter should be a string consisting of alphabetic characters, numbers, underscores, and dashes. It should not contain any spaces or special characters. The order_id must be unique within the merchant invoices/static wallets/recurrence payments. When we find an existing invoice with order_id, we return its details, a new invoice will not be created.
        :param url_callback: URL, to which the webhook will be sent after each top-up of the wallet
        :param from_referral_code: The merchant who makes the request connects to a referrer by code.
        
        
        Docs: https://doc.cryptomus.com/business/payments/creating-static
        """
        url = f"{self.__base_url}/wallet"
        params = {
            "currency": currency,
            "network": network,
            "order_id": order_id,
            "url_callback": url_callback,
            "from_referral_code": from_referral_code,
        }
        self._delete_empty_fields(params)
        self.__headers["sign"] = self.__generate_sign(params)

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, json=params)

        return GenerateStaticWallet(**response['result'])

    async def generate_qr_code_for_wallet(self, wallet_address_uuid: str) -> GenerateQrCode:
        """Generate a QR-code for the static wallet address.

        :param wallet_address_uuid: Uuid of a static wallet.

        Docs: https://doc.cryptomus.com/business/payments/qr-code-pay-form
        """
        url = f"{self.__base_url}/wallet/qr"
        params = {
            "wallet_address_uuid": wallet_address_uuid,
        }
        self.__headers["sign"] = self.__generate_sign(params)

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, json=params)

        return GenerateQrCode(**response['result'])
    
    async def generate_qr_code_for_invoice(self, merchant_payment_uuid: str) -> GenerateQrCode:
        """Generate a QR-code for the invoice address.

        :param merchant_payment_uuid: Invoice uuid.

        Docs: https://doc.cryptomus.com/business/payments/qr-code-pay-form
        """
        url = f"{self.__base_url}/payment/qr"
        params = {
            "merchant_payment_uuid": merchant_payment_uuid,
        }
        self.__headers["sign"] = self.__generate_sign(params)

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, json=params)

        return GenerateQrCode(**response['result'])
    
    async def block_static_wallet(self, uuid: Optional[str] = None, order_id: Optional[str] = None, is_force_refund: Optional[bool] = False) -> BlockStaticWallet:
        """Block static wallet.
        
        :param uuid: Uuid of a static wallet.
        :param order_id: Order ID of a static wallet.
        :param is_force_refund: Refund all incoming payments to sender’s address.
        
        Docs: https://doc.cryptomus.com/business/payments/block-wallet
        """
        url = f"{self.__base_url}/wallet/block-address"
        params = {
            "uuid": uuid,
            "order_id": order_id,
            "is_force_refund": is_force_refund,
        }
        self._delete_empty_fields(params)
        self.__headers["sign"] = self.__generate_sign(params)

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, json=params)

        return BlockStaticWallet(**response['result'])
    
    async def refund_payments_on_blocked_address(self, address: str, uuid: Optional[str] = None, order_id: Optional[str] = None) -> RefundPaymentsOnBlockedAddress:
        """Refund payments on blocked address.

        :param address: Uuid of a static wallet.
        :param uuid: Order ID of a static wallet.
        :param order_id: Refund all blocked funds to this address.

        Docs: https://doc.cryptomus.com/business/payments/refundblocked
        """
        url = f"{self.__base_url}/blocked-address-refund"
        params = {
            "address": address,
            "uuid": uuid,
            "order_id": order_id,
        }
        self._delete_empty_fields(params)
        self.__headers["sign"] = self.__generate_sign(params)

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, json=params)

        return RefundPaymentsOnBlockedAddress(**response['result'])

    async def refund(self, address: str, is_subtract: bool, uuid: Optional[str] = None, order_id: Optional[str] = None) -> bool:
        """Refund.
        
        :param address: The address to which the refund should be made.
        :param is_subtract: Whether to take a commission from the merchant's balance or from the refund amount. True - take the commission from merchant balance. False - reduce the refundable amount by the commission amount.
        :param uuid: Invoice uuid.
        :param order_id: Invoice order ID.

        Docs: https://doc.cryptomus.com/business/payments/refund
        """
        url = f"{self.__base_url}/payment/refund"
        params = {
            "address": address,
            "is_subtract": is_subtract,
            "uuid": uuid,
            "order_id": order_id,
        }
        self._delete_empty_fields(params)
        self.__headers["sign"] = self.__generate_sign(params)

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, json=params)

        return True if response['state'] == 0 else False

    async def resend_webhook(self, uuid: Optional[str] = None, order_id: Optional[str] = None) -> bool:
        """Resend webhook.
        
        :param uuid: Invoice uuid.
        :param order_id: Invoice order ID.

        Docs: https://doc.cryptomus.com/business/payments/resend-webhook
        """
        url = f"{self.__base_url}/payment/resend"
        params = {
            "uuid": uuid,
            "order_id": order_id,
        }
        self._delete_empty_fields(params)
        self.__headers["sign"] = self.__generate_sign(params)

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, json=params)

        return True if response['state'] == 0 else False
    
    async def test_webhook_payment(self, url_callback: str, currency: str, network: str, status: str, uuid: Optional[str] = None, 
                                   order_id: Optional[str] = None) -> bool:
        """Testing payment webhook.
        
        :param url_callback: Url to which webhooks with payment status will be sent.
        :param currency: Invoice currency code.
        :param network: Invoice network code.
        :param status: Payment status.
        :param uuid: uuid of the invoice.
        :param order_id: Order ID of the invoice.

        Docs: https://doc.cryptomus.com/business/payments/testing-webhook
        """
        url = f"{self.__base_url}/test-webhook/payment"
        params = {
            "url_callback": url_callback,
            "currency": currency,
            "network": network,
            "status": status,
            "uuid": uuid,
            "order_id": order_id,
        }
        self._delete_empty_fields(params)
        self.__headers["sign"] = self.__generate_sign(params)

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, json=params)

        return True if response['state'] == 0 else False
    
    async def test_webhook_wallet(self, url_callback: str, currency: str, network: str, status: str, uuid: Optional[str] = None, 
                                   order_id: Optional[str] = None) -> bool:
        """Testing wallet webhook.
        
        :param url_callback: Url to which webhooks with payment status will be sent.
        :param currency: Invoice currency code.
        :param network: Invoice network code.
        :param status: Payment status.
        :param uuid: uuid of the invoice.
        :param order_id: Order ID of the invoice.

        Docs: https://doc.cryptomus.com/business/payments/testing-webhook
        """
        url = f"{self.__base_url}/test-webhook/wallet"
        params = {
            "url_callback": url_callback,
            "currency": currency,
            "network": network,
            "status": status,
            "uuid": uuid,
            "order_id": order_id,
        }
        self._delete_empty_fields(params)
        self.__headers["sign"] = self.__generate_sign(params)

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, json=params)

        return True if response['state'] == 0 else False
    
    async def test_webhook_payout(self, url_callback: str, currency: str, network: str, status: str, uuid: Optional[str] = None, 
                                   order_id: Optional[str] = None) -> bool:
        """Testing payout webhook.
        
        :param url_callback: Url to which webhooks with payment status will be sent.
        :param currency: Payout currency code.
        :param network: Payout network code.
        :param status: Payout status.
        :param uuid: uuid of the Payout.
        :param order_id: Order ID of the Payout.

        Docs: https://doc.cryptomus.com/business/payments/testing-webhook
        """
        url = f"{self.__base_url}/test-webhook/payout"
        params = {
            "url_callback": url_callback,
            "currency": currency,
            "network": network,
            "status": status,
            "uuid": uuid,
            "order_id": order_id,
        }
        self._delete_empty_fields(params)
        self.__headers["sign"] = self.__generate_sign(params)

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, json=params)

        return True if response['state'] == 0 else False
        
    async def list_of_services(self) -> List[ListOfServices]:
        """List of services.

        Docs: https://doc.cryptomus.com/business/payments/list-of-services
        """
        url = f"{self.__base_url}/payment/services"
        self.__headers["sign"] = self.__generate_sign()

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers)

        return [ListOfServices(**service) for service in response["result"]]
    
    async def payment_history(self, date_from: Optional[str] = None, date_to: Optional[str] = None, cursor: Optional[str] = None) -> PaymentHistory:
        """Payment history.

        :param date_from: Filtering by creation date, from. Format: YYYY-MM-DD H:mm:ss
        :param date_to: Filtering by creation date, to. Format: YYYY-MM-DD H:mm:ss
        :param cursor: Cursor to page.
        
        Docs: https://doc.cryptomus.com/business/payments/payment-history
        """
        url = f"{self.__base_url}/payment/list"
        if cursor:
            url += f"?cursor={cursor}"
        params = {
            "date_from": date_from,
            "date_to": date_to,
        }
        self._delete_empty_fields(params)
        self.__headers["sign"] = self.__generate_sign(params)

        response = await self._request(self.__payment_name, self.__post_method, url, json=params, headers=self.__headers)

        return PaymentHistory(**response['result'])

    async def create_payout(self, amount: str, currency: str, order_id: str, address: str, is_subtract: bool, network: str, 
                            url_callback: Optional[str] = None, to_currency: Optional[str] = None, 
                            course_source: Optional[str] = None, from_currency: Optional[str] = None, 
                            priority: Optional[str] = None, memo: Optional[str] = None) -> Payout:
        """Creating a payout.
        
        :param amount: Payout amount.
        :param currency: Currency code for the payout. If Currency is fiat, the to_currency parameter is required.
        :param order_id: Order ID in your system. The parameter should be a string consisting of alphabetic characters, numbers, underscores, and dashes. It should not contain any spaces or special characters. The order_id must be unique within the merchant payouts. When we find an existing payout with order_id, we return its details, a new payout will not be created.
        :param address: The address of the wallet to which the withdrawal will be made.
        :param is_subtract: Defines where the withdrawal fee will be deducted. True - from your balance. False - from payout amount, the payout amount will be decreased.
        :param network: Blockchain network code. Not required when the currency/to_currency parameters is a cryptocurrency and has only one network, for example BTC.
        :param url_callback: URL to which webhooks with payout status will be sent.
        :param to_currency: Cryptocurrency code in which the payout will be made. It is used when the currency parameter is fiat.
        :param course_source: The service from which the exchange rates are taken for conversion in the invoice. The parameter is applied only if the currency is fiat, otherwise the default value is taken from the merchant's settings. Available values: Binance, BinanceP2p, Exmo, Kucoin, Garantexio.
        :param from_currency: Allows to automatically convert the withdrawal amount and use the from_currency balance. Only USDT is available.
        :param priority: The parameter for selecting the withdrawal priority. The cost of the withdrawal fee depends on the selected parameter. This parameter is applied only in case of using the BTC, ETH, POLYGON, and BSC networks. Available values: recommended, economy, high, highest.
        :param memo: Additional identifier for TON, used to specify a particular recipient or target.

        Docs: https://doc.cryptomus.com/business/payouts/creating-payout
        """
        url = f"{self.__base_url}/payout"
        params = {
            "amount": amount,
            "currency": currency,
            "order_id": order_id,
            "address": address,
            "is_subtract": is_subtract,
            "network": network,
            "url_callback": url_callback,
            "to_currency": to_currency,
            "course_source": course_source,
            "from_currency": from_currency,
            "priority": priority,
            "memo": memo,
        }
        self._delete_empty_fields(params)
        self.__headers["sign"] = self.__generate_sign(params, True)

        response = await self._request(self.__payment_name, self.__post_method, url, json=params, headers=self.__headers)

        return Payout(**response['result'])
    
    async def payout_info(self, uuid: Optional[str] = None, order_id: Optional[str] = None) -> Payout:
        """Payout information.
        
        :param uuid: Payout uuid.
        :param order_id: Payout order ID.

        Docs: https://doc.cryptomus.com/business/payouts/payout-information
        """
        url = f"{self.__base_url}/payout/info"
        params = {
            "uuid": uuid,
            "order_id": order_id,
        }
        self._delete_empty_fields(params)
        self.__headers["sign"] = self.__generate_sign(params, True)

        response = await self._request(self.__payment_name, self.__post_method, url, json=params, headers=self.__headers)

        return Payout(**response['result'])
    
    async def payout_history(self, date_from: Optional[str] = None, date_to: Optional[str] = None, cursor: Optional[str] = None) -> PayoutHistory:
        """Payout history.
        
        :param date_from: Filtering by creation date, from. Format: YYYY-MM-DD H:mm:ss.
        :param date_to: Filtering by creation date, to. Format: YYYY-MM-DD H:mm:ss.
        :param cursor: Cursor to page.

        Docs: https://doc.cryptomus.com/business/payouts/payout-history
        """
        url = f"{self.__base_url}/payout/list"
        if cursor:
            url += f"?cursor={cursor}"
        params = {
            "date_from": date_from,
            "date_to": date_to,
        }
        self._delete_empty_fields(params)
        self.__headers["sign"] = self.__generate_sign(params, True)

        response = await self._request(self.__payment_name, self.__post_method, url, json=params, headers=self.__headers)

        return PayoutHistory(**response['result'])

    async def list_of_services_payout(self) -> List[ListOfServicesPayout]:
        """List of services.
        
        Docs: https://doc.cryptomus.com/business/payouts/list-of-services
        """
        url = f"{self.__base_url}/payout/services"
        self.__headers["sign"] = self.__generate_sign(dict(), True)

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers)

        return [ListOfServicesPayout(**service) for service in response["result"]]
    
    async def transfer_to_personal_wallet(self, amount: str, currency: str) -> TransferWallet:
        """Transfer to personal wallet.
        
        :param amount: Amount to transfer.
        :param currency: Currency code. Only cryptocurrency code is allowed..

        Docs: https://doc.cryptomus.com/business/payouts/transfer-to-personal
        """
        url = f"{self.__base_url}/transfer/to-personal"
        params = {
            "amount": amount,
            "currency": currency,
        }
        self.__headers["sign"] = self.__generate_sign(params, True)

        response = await self._request(self.__payment_name, self.__post_method, url, json=params, headers=self.__headers)

        return TransferWallet(**response['result'])
    


    async def transfer_to_business_wallet(self, amount: str, currency: str) -> TransferWallet:
        """Transfer to business wallet.
        
        :param amount: Amount to transfer.
        :param currency: Currency code. Only cryptocurrency code is allowed..

        Docs: https://doc.cryptomus.com/business/payouts/transfer-to-personal
        """
        url = f"{self.__base_url}/transfer/to-business"
        params = {
            "amount": amount,
            "currency": currency,
        }
        self.__headers["sign"] = self.__generate_sign(params, True)

        response = await self._request(self.__payment_name, self.__post_method, url, json=params, headers=self.__headers)

        return TransferWallet(**response['result'])
    
    async def creating_recurring_payment(self, amount: str, currency: str, name: str, period: str, to_currency: Optional[str] = None,
                                         order_id: Optional[str] = None, url_callback: Optional[str] = None, discount_days: Optional[str] = None,
                                         discount_amount: Optional[str] = None, additional_data: Optional[str] = None) -> RecurringPayment:
        """Creating recurring payment.
        
        :param amount: Recurring payment amount.
        :param currency: Currency code.
        :param name: Recurring payment name.
        :param period: Recurring payment period. Available: weekly, monthly, three_month.
        :param to_currency: Currency code for accepting payments. The parameter is used to specify the target currency for converting the recurrent payment amount. For example, to create an recurrent payment for 20 USD in bitcoin: amount: 20, currency: USD, to_currency: BTC. The API will convert 20 USD amount to its equivalent in BTC based on the current exchange rate and the user will pay in BTC. The to_currency should always be the cryptocurrency code, not a fiat currency code.
        :param order_id: Order ID in your system.
        :param url_callback: Url to which webhooks with payment status will be sent.
        :param discount_days: Discount period days (required with 'discount_amount').
        :param discount_amount: Discount amount (required with 'discount_days').Here the amount in the currency of the parameter ‘currency’.
        :param additional_data: Additional recurring payment details.

        Docs: https://doc.cryptomus.com/business/recurring/creating
        """
        url = f"{self.__base_url}/recurrence/create"
        params = {
            "amount": amount,
            "currency": currency,
            "name": name,
            "period": period,
            "to_currency": to_currency,
            "order_id": order_id,
            "url_callback": url_callback,
            "discount_days": discount_days,
            "discount_amount": discount_amount,
            "additional_data": additional_data,
        }
        self._delete_empty_fields(params)
        self.__headers["sign"] = self.__generate_sign(params)

        response = await self._request(self.__payment_name, self.__post_method, url, json=params, headers=self.__headers)

        return RecurringPayment(**response['result'])

    async def recurring_payment_info(self, uuid: Optional[str] = None, order_id: Optional[str] = None) -> RecurringPayment:
        """Payment information.
        
        :param uuid: Recurrence uuid.
        :param order_id: Recurrence order ID.

        Docs: https://doc.cryptomus.com/business/recurring/info"""
        url = f"{self.__base_url}/recurrence/info"
        params = {
            "uuid": uuid,
            "order_id": order_id,
        }
        self._delete_empty_fields(params)
        self.__headers["sign"] = self.__generate_sign(params)

        response = await self._request(self.__payment_name, self.__post_method, url, json=params, headers=self.__headers)

        return RecurringPayment(**response['result'])

    async def list_of_recurring_payments(self, cursor: Optional[str] = None) -> ListOfRecurringPayments:
        """List of recurring payments.
        
        Docs: https://doc.cryptomus.com/business/recurring/list
        """
        url = f"{self.__base_url}/recurrence/list"
        if cursor:
            url += f"?cursor={cursor}"
        self.__headers["sign"] = self.__generate_sign()

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers)

        return ListOfRecurringPayments(**response['result'])

    async def cancel_recurring_payment(self, uuid: Optional[str] = None, order_id: Optional[str] = None) -> RecurringPayment:
        """Cancel recurring payment
        
        :param uuid: Recurrence uuid.
        :param order_id: Order ID in your system.

        Docs: https://doc.cryptomus.com/business/recurring/cancel
        """
        url = f"{self.__base_url}/recurrence/cancel"
        params = {
            "uuid": uuid,
            "order_id": order_id,
        }
        self.__headers["sign"] = self.__generate_sign(params)

        response = await self._request(self.__payment_name, self.__post_method, url, json=params, headers=self.__headers)

        return RecurringPayment(**response['result'])

    async def exchange_rates_list(self, currency: str) -> List[ExchangeRatesList]:
        """Exchange rates list.
        
        Docs: https://doc.cryptomus.com/business/exchange-rates/list
        """
        url = f"{self.__base_url}/exchange-rate/{currency}/list"
        self.__headers["sign"] = self.__generate_sign()

        response = await self._request(self.__payment_name, self.__get_method, url, headers=self.__headers)

        return [ExchangeRatesList(**rate) for rate in response["result"]]

    async def list_of_discounts(self) -> List[Discount]:
        """List of discounts.
        
        Docs: https://doc.cryptomus.com/business/discount/list
        """
        url = f"{self.__base_url}/payment/discount/list"
        self.__headers["sign"] = self.__generate_sign()

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers)

        return [Discount(**discount) for discount in response["result"]]
    
    async def set_discount_to_payment_method(self, currency: str, network: str, discount_percent: int) -> Discount:
        """Set discount to payment method. 
        
        :param currency: Currency code.
        :param network: Blockchain network code.
        :param discount_percent: Discount percent.

        Docs: https://doc.cryptomus.com/business/discount/set
        """
        url = f"{self.__base_url}/payment/discount/set"
        params = {
            "currency": currency,
            "network": network,
            "discount_percent": discount_percent,
        }
        self.__headers["sign"] = self.__generate_sign(params)

        response = await self._request(self.__payment_name, self.__post_method, url, json=params, headers=self.__headers)

        return Discount(**response["result"])


import ssl
import certifi
from typing import Optional
from aiohttp import ClientSession, TCPConnector, ClientResponse
from .exceptions.exceptions import BadRequest, RequestError


class RequestsClient:

    def __init__(self) -> None:
        self._session: Optional[ClientSession] = None

    def _getsession(self) -> ClientSession:

        if isinstance(self._session, ClientSession) and not self._session.closed:
            return self._session

        ssl_context = ssl.create_default_context(cafile=certifi.where())
        connector = TCPConnector(ssl=ssl_context)

        self._session = ClientSession(connector=connector)

        return self._session
    
    def _delete_empty_fields(self, params: dict) -> None:
        for key, value in params.copy().items():
            if value is None:
                params.pop(key)

    async def _request_for_authorize_yoomoney(self, method: str, url: str, **kwargs) -> ClientResponse:
        session = self._getsession()
        async with session.request(method, url, **kwargs) as response:
            await self._session.close()
            return response
    
    async def _request(self, payment: str, method: str, url: str, **kwargs) -> dict:
        session = self._getsession()

        async with session.request(method, url, **kwargs) as response:
            await self._session.close()
            if response.status in [200, 201]:
                if payment in ["ruKassa"]:
                    response = await response.json(content_type="text/html")
                elif payment in ['payok']:
                    response = await response.json(content_type="text/plain")
                elif payment == "yoomoney_quick-pay":
                    response = response.url
                else:
                    response = await response.json()
            else:
                try:
                    response_json = await response.json()
                    if response_json:
                        return await self._checkexception(payment, response_json)
                    else:
                        raise RequestError(
                            f"{payment}. Response status: {response.status}. Text: {await response.text()}"
                        )
                except:
                    raise RequestError(
                        f"{payment}. Response status: {response.status}. Text: {await response.text()}"
                    )
            return response

    async def _checkexception(self, payment: str, response: dict) -> dict:
        if payment == "aaio":
            if response["type"] == "error":
                raise BadRequest("[AAIO] " + response["message"])
        elif payment == "crystalPay":
            if response["error"]:
                raise BadRequest("[CrystalPay] " + response["errors"][0])
        elif payment == "cryptoBot":
            if not response["ok"]:
                raise BadRequest("[CryptoBot] " + response["error"]["name"])
        elif payment == "lolz":
            if response.get("error"):
                raise BadRequest("[Lolzteam Market] " + response["error_description"])
            if response.get("errors"):
                raise BadRequest("[Lolzteam Market] " + response["errors"][0])
        elif payment == "ruKassa":
            if response.get("error"):
                raise BadRequest("[RuKassa] " + response["message"])
        elif payment == "freeKassa":
            if response["type"] == "error" and not response.get("description"):
                raise BadRequest("[FreeKassa] " + response["message"])
        elif payment == "cryptomus":
            if response.get('state') != 0:
                if response.get("errors"):
                    raise BadRequest("[Cryptomus] " + str(response.get("errors")))
                else:
                    raise BadRequest("[Cryptomus] " + str(response.get("message")))
            else:
                raise BadRequest("[Cryptomus] " + response.get("message"))
        elif payment == "xrocket":
            if not response.get("success"):
                text = f"[XRocket] {response.get('message')}"
                if response.get("errors"):
                    text += ": \n"
                    for error in response.get("errors"):
                        text += f"Property: {error['property']} \nError: {error['error']}"
                raise BadRequest(text)
            else:
                raise BadRequest(f"[XRocket] Status code: {response.get('statusCode')}. Message: " + response.get("message"))
        elif payment == "yoomoney":
            if response.get("error"):
                raise BadRequest("[YooMoney] " + response.get("error_description") + ". Error code: " + response['error'])
        else:
            # payok
            if response.get("status") and response.pop("status") == "error":
                raise BadRequest("[PayOK] " + response.get("text", response.get("error_text")) + ". Error code: " + response['error_code'])

        return response



from ..requests import RequestsClient
from typing import Optional, Union
from .models import Order, OrderInfo, ExchangeRate
from urllib.parse import urlencode
import json

class AsyncPlatega(RequestsClient):
    API_HOST: str = "https://platega.io/"

    def __init__(self, merchant_id: str, secret_key: str) -> None:
        """
        Initialize Platega API client
        :param merchant_id: Your Platega merchant ID.
        :param secret_key: Your Platega secret key.
        """
        super().__init__()
        self.__merchant_id = merchant_id
        self.__secret_key = secret_key
        self.__headers = {
            "X-MerchantId": merchant_id,
            "X-Secret": secret_key,
            "Content-Type": "application/json",
        }
        self.__base_url = "https://app.platega.io"
        self.__get_method = "GET"
        self.__post_method = "POST"
        self.__payment_name = "platega"
        self.check_values()

    def check_values(self):
        if not self.__secret_key or not self.__merchant_id:
            raise ValueError('No SecretKey or MerchantID specified')
        
    async def create_order(
        self,
        payment_method: str,
        amount: Union[int, float],
        currency: str,
        description: str,
        return_url: Optional[str] = None,
        failed_url: Optional[str] = None,
        payload: Optional[str] = None
    ) -> Order:
        """Creates a transaction and returns payment details. The transaction ID is generated automatically by the system—do not pass the id field in the request.
        
        :param payment_method: Payment method number (for example, 2 for QR SBP). 2 - SBP QR; 11 - Card acquiring; 12 - International acquiringInternational card payments; 13 - Cryptocurrency.
        :param amount: Payment amount.
        :param currency: Payment currency (for example, RUB).
        :param description: The purpose (description) of the payment, please always indicate it if possible.
        :param return_url: Redirect upon successful payment.
        :param failed_url: Redirect for unsuccessful payment.
        :param payload: Additional information for initialization on your system.
        
        Docs: https://docs.platega.io/%D1%81%D0%BE%D0%B7%D0%B4%D0%B0%D0%BD%D0%B8%D0%B5-%D1%81%D1%81%D1%8B%D0%BB%D0%BA%D0%B8-%D0%BD%D0%B0-%D0%BE%D0%BF%D0%BB%D0%B0%D1%82%D1%83-27021914e0
        """
        params = {
            "paymentMethod": payment_method,
            "paymentDetails": {
                "amount": amount,
                "currency": currency
            },
            "description": description,
            "return": return_url,
            "failedUrl": failed_url,
            "payload": payload,
        }
        self._delete_empty_fields(params)
        url = f"{self.__base_url}/transaction/process"

        response = await self._request(
            self.__payment_name,
            self.__post_method,
            url,
            headers=self.__headers,
            data=json.dumps(params),
        )
        return Order(**response)
    
    async def get_order(self, order_id: str) -> OrderInfo:
        """Returns the status and details of the transaction.

        :param order_id: Transaction ID (UUID).
        
        Docs: https://docs.platega.io/%D0%BF%D1%80%D0%BE%D0%B2%D0%B5%D1%80%D0%BA%D0%B0-%D1%81%D1%82%D0%B0%D1%82%D1%83%D1%81%D0%B0-%D0%BE%D0%BF%D0%BB%D0%B0%D1%82%D1%8B-%D0%BF%D0%BB%D0%B0%D1%82%D0%B5%D0%B6%D0%B0-22645077e0
        """
        url = f"{self.__base_url}/transaction/{order_id}"

        response = await self._request(
            self.__payment_name,
            self.__get_method,
            url,
            headers=self.__headers,
        )
        return OrderInfo(**response)
    
    async def get_rates(self, payment_method: str, currency_from: str, currency_to: str) -> ExchangeRate:
        """Returns the current exchange rate for the specified payment method and currencies.
        
        :param payment_method: Payment method number (for example, 2 for QR SBP). 2 - SBP QR; 11 - Card acquiring; 12 - International acquiringInternational card payments; 13 - Cryptocurrency.
        :param currency_from: Source currency (e.g. RUB).
        :param currency_to: Target currency (e.g. USDT).
        
        Docs: https://docs.platega.io/%D0%BF%D0%BE%D0%BB%D1%83%D1%87%D0%B5%D0%BD%D0%B8%D0%B5-%D0%BA%D1%83%D1%80%D1%81%D0%BE%D0%B2-%D0%BF%D0%BE-%D0%BF%D0%BB%D0%B0%D1%82%D0%B5%D0%B6%D0%BD%D0%BE%D0%BC%D1%83-%D0%BC%D0%B5%D1%82%D0%BE%D0%B4%D1%83-22645078e0
        """
        params = {
            "merchantId": self.__merchant_id,
            "paymentMethod": payment_method,
            "currencyFrom": currency_from,
            "currencyTo": currency_to,
        }
        url = f"{self.__base_url}/rates/payment_method_rate"

        response = await self._request(
            self.__payment_name,
            self.__post_method,
            url,
            headers=self.__headers,
            data=urlencode(params),
        )
        return ExchangeRate(**response)
    
    async def get_orders(self, date_from: str, date_to: str, page: str, size: str):
        """Method for receiving conversions.
        
        :param date_from: Date from. Example: 2025-01-01T00%3A00%3A00Z.
        :param date_to: Date to. Example: 2025-11-13T23%3A59%3A59Z.
        :param page: Page number. Example: 1.
        :param size: Size. Example: 20.
        
        Docs: https://docs.platega.io/%D0%BF%D0%BE%D0%BB%D1%83%D1%87%D0%B5%D0%BD%D0%B8%D0%B5-%D0%BA%D1%83%D1%80%D1%81%D0%BE%D0%B2-%D0%BF%D0%BE-%D0%BF%D0%BB%D0%B0%D1%82%D0%B5%D0%B6%D0%BD%D0%BE%D0%BC%D1%83-%D0%BC%D0%B5%D1%82%D0%BE%D0%B4%D1%83-22645078e0
        """
        params = {
            "from": date_from,
            "to": date_to,
            "page": page,
            "size": size,
        }
        url = f"{self.__base_url}/transaction/balance-unlock-operations"
        
        new_headers = self.__headers
        new_headers['accept'] = "text/plain"
        
        response = await self._request(
            self.__payment_name,
            self.__post_method,
            url,
            headers=new_headers,
            data=urlencode(params),
        )
        return response
        


from ..requests import RequestsClient
from typing import Optional, Union
from .models import User, Payments, Invoice, Invoices
from ..exceptions import MissingScopeError, IncorrectTokenError, UnexpectedError

import json
import random
import math
import time
import secrets
import base64
from urllib.parse import urlencode


class AsyncLolzteamMarketPayment(RequestsClient):
    API_HOST: str = "https://lzt.market"

    def __init__(self, token: str) -> None:
        """
        Initialize LolzteamMarket API client
        :param token: Your Lolzteam Token
        """
        super().__init__()
        self.__token = token
        try:
            jwt_payload = json.loads(
                base64.b64decode(token.split(".")[1] + "==").decode("utf-8")
            )
        except IndexError:
            raise IncorrectTokenError(
                "You have specified an incorrect token."
            )
        except Exception as error:
            raise UnexpectedError(error)
            
        self.__user_id = jwt_payload["sub"]
        if jwt_payload.get("scope"):
            if "market" not in jwt_payload["scope"]:
                raise MissingScopeError(
                    '"Market" scope is not provided in your token. You need to recreate token with "Market" scope.'
                )
        self.__headers = {
            "Authorization": f"Bearer {self.__token}",
            "Accept": "application/json",
        }
        self.__base_url = "https://prod-api.lzt.market"
        self.__get_method = "GET"
        self.__post_method = "POST"
        self.__payment_name = "lolz"

    async def get_me(self) -> User:
        """Get info about your account on Zelenka (Lolzteam).

        Docs: https://lzt-market.readme.io/reference/marketprofilesettingsgetinfo"""

        url = f"{self.__base_url}/me"

        response = await self._request(
            self.__payment_name, self.__get_method, url, headers=self.__headers
        )

        return User(**response["user"])

    def __get_random_string(self):
        return f"{time.time()}_{secrets.token_hex(random.randint(5, 10))}"

    async def create_invoice(
        self, 
        amount: Union[int, float], 
        payment_id: str, 
        comment: str, 
        url_success: str, 
        merchant_id: str, 
        currency: Optional[str] = "rub", 
        url_callback: Optional[str] = None, 
        lifetime: Optional[int] = 3600, 
        additional_data: Optional[str] = None,
        is_test: Optional[bool] = False,
        required_telegram_id: Optional[int] = None,
        required_telegram_username: Optional[str] = None
    ) -> Invoice:
        """Create invoice.
        
        :param amount: Invoice amount.
        :param payment_id: Payment ID in your system (must be unique within the merchant / invoices).
        :param comment: Comment to the invoice.
        :param url_success: URL to redirect to after successful payment.
        :param merchant_id: Merchant ID.
        :param currency: Optional. Currency that will be used to create the invoice. Defaults to 'rub'.
        :param url_callback: Optional. Callback url.
        :param lifetime: Optional. Invoice lifetime.
        :param additional_data: Optional. Additional information for you.
        :param is_test: Optional. Create a test invoice.
        :param required_telegram_id: Optional. Telegram User ID for which the invoice was created.
        :param required_telegram_username: Optional. Telegram Username (including @) for which the invoice was created (if any).
        
        Docs: https://lzt-market.readme.io/reference/paymentsinvoicecreate
        """
        params = {
            "amount": amount,
            "payment_id": payment_id,
            "comment": comment,
            "url_success": url_success,
            "merchant_id": merchant_id,
            "currency": currency,
            "url_callback": url_callback,
            "lifetime": lifetime,
            "additional_data": additional_data,
            "is_test": is_test,
            "required_telegram_id": required_telegram_id,
            "required_telegram_username": required_telegram_username,
        }
        self._delete_empty_fields(params)
        url = f"{self.__base_url}/invoice"

        response = await self._request(
            self.__payment_name,
            self.__post_method,
            url,
            headers=self.__headers,
            json=params,
        )
        return Invoice(**response['invoice'])
    
    async def get_invoice(
        self, 
        invoice_id: Optional[str] = None, 
        payment_id: Optional[str] = None
    ) -> Invoice:
        """Get invoice.
        
        :param invoice_id: Optional. Invoice ID.
        :param payment_id: Optional. Payment ID.
        
        Docs: https://lzt-market.readme.io/reference/paymentsinvoiceget
        """
        
        params = {
            "invoice_id": invoice_id,
            "payment_id": payment_id,
        }
        
        self._delete_empty_fields(params)
        
        url = f"{self.__base_url}/invoice?{urlencode(params)}"

        response = await self._request(
            self.__payment_name,
            self.__get_method,
            url,
            headers=self.__headers,
        )
        return Invoice(**response['invoice'])
    
    async def get_invoice_list(
        self, 
        page: Optional[int] = 1,
        currency: Optional[str] = None,
        status: Optional[str] = None,
        amount: Optional[Union[float, int]] = None,
        merchant_id: Optional[int] = None,
    ) -> Invoices:
        """Get invoice list.
        
        :param page: Optional. The number of the page to display results from.
        :param currency: Optional. Currency of the created invoice.
        :param status: Optional. Status of the invoice.
        :param amount: Optional. Invoice amount.
        :param merchant_id: Optional. Merchant ID.
        
        Docs: https://lzt-market.readme.io/reference/paymentsinvoicelist
        """

        url = f"{self.__base_url}/invoice/list"

        params = {
            "page": page,
            "currency": currency,
            "status": status,
            "amount": amount,
            "merchant_id": merchant_id,
        }

        self._delete_empty_fields(params)

        response = await self._request(
            self.__payment_name,
            self.__get_method,
            url,
            headers=self.__headers,
            params=params,
        )
        return Invoices(**response)
    
    def get_payment_link(
        self,
        amount: Union[int, float],
        comment: Optional[str] = None,
        is_hold: Optional[bool] = False,
        is_amount_ceiling: Optional[bool] = False,
    ) -> str:
        """Get a link to transfer funds to the Lolzteam market. To accept payments, it is recommended to use the create_invoice() method.

        :param amount: Amount to transfer
        :param comment: Comment on the translation. If not specified: a random unique set of characters is generated.
        :param is_hold: If True: The page will have funds holding enabled by default. The user will be able to turn it off. *If you use this link for payment: ask your users not to enable hold!!! Defaults to False
        :param is_amount_ceiling: If True: The transfer amount will be rounded up. Defaults to False

        :return: Link to transfer (String)"""

        if is_amount_ceiling:
            amount = math.ceil(amount)

        if not comment:
            comment = self.__get_random_string()

        return f"https://lzt.market/balance/transfer?user_id={self.__user_id}&hold={int(is_hold)}&amount={amount}&comment={comment}"

    async def get_history_payments(
        self,
        operation_type: Optional[str] = None,
        pmin: Optional[int] = None,
        pmax: Optional[int] = None,
        page: Optional[int] = 1,
        operation_id_lt: Optional[int] = None,
        receiver: Optional[str] = None,
        sender: Optional[str] = None,
        startDate: Optional[str] = None,
        endDate: Optional[str] = None,
        wallet: Optional[str] = None,
        comment: Optional[str] = None,
        is_hold: Optional[bool] = None,
        show_payment_stats: Optional[bool] = None,
    ) -> Payments:
        """Displays list of your payments.

        Docs: https://lzt-market.readme.io/reference/paymentslisthistory

        :param operation_type: Optional. Type of operation. It can only be: "income", "cost", "refilled_balance", "withdrawal_balance", "paid_item", "sold_item", "money_transfer", "receiving_money", "internal_purchase", "claim_hold".
        :param pmin: Optional. Minimal price of account (Inclusive).
        :param pmax: Optional. Maximum price of account (Inclusive).
        :param page: Optional. The number of the page to display results from.
        :param operation_id_lt: Optional. ID of the operation from which the result begins.
        :param receiver: Optional. Username of user, which receive money from you.
        :param sender: Optional. Username of user, which sent money to you.
        :param startDate: Optional. Start date of operation (RFC 3339 date format).
        :param endDate: Optional. End date of operation (RFC 3339 date format).
        :param wallet: Optional. Wallet, which used for money payouts.
        :param comment: Optional. Comment for money transfers.
        :param is_hold: Optional. Display hold operations.
        :param show_payment_stats: Optional. Display payment stats for selected period (outgoing value, incoming value).
        """

        url = f"{self.__base_url}/user/{self.__user_id}/payments"

        if is_hold:
            is_hold = int(is_hold)
        if show_payment_stats:
            show_payment_stats = int(show_payment_stats)

        params = {
            "type": operation_type,
            "pmin": pmin,
            "pmax": pmax,
            "page": page,
            "operation_id_lt": operation_id_lt,
            "receiver": receiver,
            "sender": sender,
            "startDate": startDate,
            "endDate": endDate,
            "wallet": wallet,
            "comment": comment,
            "is_hold": is_hold,
            "show_payment_stats": show_payment_stats,
        }

        self._delete_empty_fields(params)

        response = await self._request(
            self.__payment_name,
            self.__get_method,
            url,
            headers=self.__headers,
            params=params,
        )
        if type(response.get("payments")) is list and len(response.get("payments")) == 0:
            response["payments"] = {}
        return Payments(**response)

    async def check_status_payment(self, pay_amount: int, comment: str) -> bool:
        """Displays whether the transfer is paid or not.

        :param pay_amount: The amount indicated in the transaction.
        :param comment: Comment indicated in the transaction.

        :return: True if payment has been received. Otherwise False"""
        payments = (
            await self.get_history_payments(
                operation_type="receiving_money",
                comment=comment,
                pmin=pay_amount,
                pmax=pay_amount,
            )
        ).payments
        if payments.values():
            return True
        return False



class BadRequest(Exception):
    pass


class RequestError(Exception):
    pass


class MissingScopeError(Exception):
    pass


class IncorrectTokenError(Exception):
    pass


class UnexpectedError(Exception):
    pass


class InvalidRequest(Exception):
    pass


class UnauthorizedClient(Exception):
    pass


class InvalidGrant(Exception):
    pass


class EmptyToken(Exception):
    pass


from ..requests import RequestsClient
from typing import Optional, List
from .models import AppInfo, Transfer, Withdrawal, WithdrawalFees, MultiCheque, MultiChequesList, Invoice, InvoicesList, Currency, \
                    Subscription, SubscriptionsList, SubscriptionCheck, Subscriptions
from urllib.parse import urlencode


class AsyncXRocket(RequestsClient):
    API_HOST: str = "https://t.me/tonRocketBot"

    def __init__(self, apiKey: str) -> None:
        """
        Initialize XRocket API client
        :param apiKey: Your API key
        """
        super().__init__()
        self.__api_key = apiKey
        self.__headers = {
            "Content-Type": "application/json",
            "Rocket-Pay-Key": self.__api_key,
        }
        self.__base_url = "https://pay.xrocket.tg"
        self.__post_method = "POST"
        self.__get_method = "GET"
        self.__put_method = "PUT"
        self.__delete_method = "DELETE"
        self.__payment_name = "xrocket"
        self.check_values()

    def check_values(self):
        if not self.__api_key:
            raise ValueError('No API key specified')
        
    async def get_app_info(self) -> AppInfo:
        """Returns information about your application.
        
        Docs: https://pay.xrocket.tg/api/#/app/AppsController_getAppInfo"""
        url = f"{self.__base_url}/app/info"
        response = await self._request(self.__payment_name, self.__get_method, url, headers=self.__headers)
        return AppInfo(**response['data'])

    async def transfer(self, tgUserId: int, currency: str, amount: float, transferId: str, description: Optional[str] = "") -> Transfer:
        """Make transfer of funds to another user.
        
        :param tgUserId: Telegram user ID. If we dont have this user in DB, we will fail transaction with error: 400 - User not found.
        :param currency: Currency of transfer, info with function get_available_currencies().
        :param amount: Transfer amount. 9 decimal places, others cut off.
        :param transferId: Unique transfer ID in your system to prevent double spends.
        :description: Transfer description.

        Docs: https://pay.xrocket.tg/api/#/app/AppsController_transfer"""
        url = f"{self.__base_url}/app/transfer"
        params = {
            "tgUserId": tgUserId,
            "currency": currency,
            "amount": amount,
            "transferId": transferId,
            "description": description,
        }
        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, json=params)
        return Transfer(**response['data'])
    
    async def withdrawal(self, network: str, address: str, currency: str, amount: float, withdrawalId: str, comment: Optional[str] = "") -> Withdrawal:
        """Make withdrawal of funds to external wallet.
        
        :param network: Network code.
        :param address: Withdrawal address.
        :param currency: Currency code.
        :param amount: Withdrawal amount. 9 decimal places, others cut off.
        :param withdrawalId: Unique withdrawal ID in your system to prevent double spends.
        :param comment: Withdrawal comment.

        Docs: https://pay.xrocket.tg/api/#/app/AppsController_withdrawal"""
        url = f"{self.__base_url}/app/withdrawal"
        params = {
            "network": network,
            "address": address,
            "currency": currency,
            "amount": amount,
            "withdrawalId": withdrawalId,
            "comment": comment,
        }
        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, json=params)
        return Withdrawal(**response['data'])
    
    async def withdrawal_status(self, withdrawalId: str) -> Withdrawal:
        """Returns withdrawal status.
        
        Docs: https://pay.xrocket.tg/api/#/app/AppsController_getWithdrawalStatus"""
        url = f"{self.__base_url}/app/withdrawal/status/{withdrawalId}"
        response = await self._request(self.__payment_name, self.__get_method, url, headers=self.__headers)
        return Withdrawal(**response['data'])
    
    async def withdrawal_fees(self, currency: Optional[str] = None) -> List[WithdrawalFees]:
        """Returns withdrawal fees.
        
        Docs: https://pay.xrocket.tg/api/#/app/AppsController_getWithdrawalFees"""
        url = f"{self.__base_url}/app/withdrawal/fees"
        params = {
            "currency": currency,
        }
        self._delete_empty_fields(params)
        response = await self._request(self.__payment_name, self.__get_method, url, headers=self.__headers, json=params)
        return [WithdrawalFees(**withdrawalFee) for withdrawalFee in response['data']]
    
    async def create_multi_cheque(self, currency: str, chequePerUser: float, usersNumber: int, refProgram: int, password: Optional[str] = None,
                                  description: Optional[str] = None, sendNotifications: Optional[bool] = True, enableCaptcha: Optional[bool] = True,
                                  telegramResourcesIds: Optional[List[str]] = [], forPremium: Optional[bool] = False, linkedWallet: Optional[bool] = False,
                                  disabledLanguages: Optional[List[str]] = [], enabledCountries: Optional[List[str]] = []) -> MultiCheque:
        """Create multi-cheque.

        :param currency: Currency of transfer, info with function get_available_currencies().
        :param chequePerUser: Cheque amount for one user. 9 decimal places, others cut off.
        :param usersNumber: Number of users to save multicheque. 0 decimal places.
        :param refProgram: Referral program percentage (%). 0 decimal places.
        :param password: Password for cheque.
        :param description: Description for cheque.
        :param sendNotifications: Send notifications about activations.
        :param enableCaptcha: Enable captcha.
        :param telegramResourcesIds: IDs of telegram resources (groups, channels, private groups).
        :param forPremium: Only users with Telegram Premium can activate this cheque.
        :param linkedWallet: Only users with linked wallet can activate this cheque.
        :param disabledLanguages: Disable languages.
        :param enabledCountries: Enabled countries (AsyncPayments.xrocket.models.CountriesName).
        
        Docs: https://pay.xrocket.tg/api/#/multi-cheque/ChequesController_createCheque"""
        url = f"{self.__base_url}/multi-cheque"
        params = {
            "currency": currency,
            "chequePerUser": chequePerUser,
            "usersNumber": usersNumber,
            "refProgram": refProgram,
            "password": password,
            "description": description,
            "sendNotifications": sendNotifications,
            "enableCaptcha": enableCaptcha,
            "telegramResourcesIds": telegramResourcesIds,
            "forPremium": forPremium,
            "linkedWallet": linkedWallet,
            "disabledLanguages": disabledLanguages,
            "enabledCountries": enabledCountries,
        }
        self._delete_empty_fields(params)
        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, json=params)
        return MultiCheque(**response['data'])
    
    async def multi_cheques_list(self, limit: Optional[int] = 100, offset: Optional[int] = 0) -> MultiChequesList:
        """Get list of multi-cheques.
        
        :param limit: Limit of cheques.
        :param offset: Offset.

        Docs: https://pay.xrocket.tg/api/#/multi-cheque/ChequesController_getCheques"""
        params = {
            "limit": limit,
            "offset": offset,
        }
        url = f"{self.__base_url}/multi-cheque?{urlencode(params)}"
        response = await self._request(self.__payment_name, self.__get_method, url, headers=self.__headers)
        return MultiChequesList(**response['data'])
    
    async def get_multi_cheque_info(self, cheque_id: int) -> MultiCheque:
        """Get multi-cheque info.
        
        :param cheque_id: ID of cheque.

        Docs: https://pay.xrocket.tg/api/#/multi-cheque/ChequesController_getCheque"""
        url = f"{self.__base_url}/multi-cheque/{cheque_id}"
        response = await self._request(self.__payment_name, self.__get_method, url, headers=self.__headers)
        return MultiCheque(**response['data'])
    
    async def edit_multi_cheque(self, cheque_id: int, password: Optional[str] = None, description: Optional[str] = None,
                                sendNotifications: Optional[bool] = None, enableCaptcha: Optional[bool] = None, 
                                telegramResourcesIds: Optional[List[str]] = None, forPremium: Optional[bool] = None,
                                linkedWallet: Optional[bool] = None, disabledLanguages: Optional[List[str]] = None,
                                enabledCountries: Optional[List[str]] = None) -> MultiCheque:
        """Edit multi-cheque.
        
        :param cheque_id: ID of cheque.
        :param password: Password for cheque.
        :param description: Description for cheque.
        :param sendNotifications: Send notifications about activations.
        :param enableCaptcha: Enable captcha.
        :param telegramResourcesIds: IDs of telegram resources (groups, channels, private groups).
        :param forPremium: Only users with Telegram Premium can activate this cheque.
        :param linkedWallet: Only users with linked wallet can activate this cheque.
        :param disabledLanguages: Disable languages.
        :param enabledCountries: Enabled countries (AsyncPayments.xrocket.models.CountriesName).
        
        
        Docs: https://pay.xrocket.tg/api/#/multi-cheque/ChequesController_editCheque"""
        params = {
            "password": password,
            "description": description,
            "sendNotifications": sendNotifications,
            "enableCaptcha": enableCaptcha,
            "telegramResourcesIds": telegramResourcesIds,
            "forPremium": forPremium,
            "linkedWallet": linkedWallet,
            "disabledLanguages": disabledLanguages,
            "enabledCountries": enabledCountries,
        }
        self._delete_empty_fields(params)
        url = f"{self.__base_url}/multi-cheque/{cheque_id}"
        response = await self._request(self.__payment_name, self.__put_method, url, headers=self.__headers, json=params)
        return MultiCheque(**response['data'])
    
    async def delete_multi_cheque(self, cheque_id: int) -> bool:
        """Delete multi-cheque.
        
        :param cheque_id: ID of cheque.

        Docs: https://pay.xrocket.tg/api/#/multi-cheque/ChequesController_deleteCheque"""
        url = f"{self.__base_url}/multi-cheque/{cheque_id}"
        response = await self._request(self.__payment_name, self.__delete_method, url, headers=self.__headers)
        return True if response['success'] else False
    
    async def create_invoice(self, numPayments: int, currency: str, amount: Optional[float] = None, minPayment: Optional[int] = None, 
                             description: Optional[str] = None, hiddenMessage: Optional[str] = None, commentsEnabled: Optional[bool] = None,
                             callbackUrl: Optional[str] = None, payload: Optional[str] = None, expiredIn: Optional[int] = None) -> Invoice:
        """Create invoice.
        
        :param numPayments: Num payments for invoice.
        :param currency: Currency of transfer, info with function get_available_currencies().
        :param amount: Invoice amount. 9 decimal places, others cut off.
        :param minPayment: Min payment only for multi invoice if invoice amount is None.
        :param description: Description for invoice.
        :param hiddenMessage: Hidden message after invoice is paid.
        :param commentsEnabled: Allow comments.
        :param callbackUrl: Url for Return button after invoice is paid.
        :param payload: Any data. Invisible to user, will be returned in callback.
        :param expiredIn: Invoice expire time in seconds, max 1 day, 0 - None expired.

        Docs: https://pay.xrocket.tg/api/#/tg-invoices/InvoicesController_createInvoice"""
        params = {
            "numPayments": numPayments,
            "currency": currency,
            "amount": amount,
            "minPayment": minPayment,
            "description": description,
            "hiddenMessage": hiddenMessage,
            "commentsEnabled": commentsEnabled,
            "callbackUrl": callbackUrl,
            "payload": payload,
            "expiredIn": expiredIn,
        }
        self._delete_empty_fields(params)
        url = f"{self.__base_url}/tg-invoices"
        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, json=params)
        return Invoice(**response['data'])
    
    async def get_list_invoices(self, limit: Optional[int] = 100, offset: Optional[int] = 0) -> InvoicesList:
        """Get list of invoices.
        
        :param limit: Limit of invoices.
        :param offset: Offset.

        Docs: https://pay.xrocket.tg/api/#/tg-invoices/InvoicesController_getInvoices"""
        params = {
            "limit": limit,
            "offset": offset,
        }
        url = f"{self.__base_url}/tg-invoices?{urlencode(params)}"
        response = await self._request(self.__payment_name, self.__get_method, url, headers=self.__headers)
        return InvoicesList(**response['data'])
    
    async def get_invoice_info(self, invoice_id: str) -> Invoice:
        """Get invoice info.
        
        :param invoice_id: ID of invoice.
        
        Docs: https://pay.xrocket.tg/api/#/tg-invoices/InvoicesController_getInvoice"""
        url = f"{self.__base_url}/tg-invoices/{invoice_id}"
        response = await self._request(self.__payment_name, self.__get_method, url, headers=self.__headers)
        return Invoice(**response['data'])
    
    async def delete_invoice(self, invoice_id: str) -> bool:
        """Delete invoice.
        
        :param invoice_id: ID of invoice.

        Docs: https://pay.xrocket.tg/api/#/tg-invoices/InvoicesController_deleteInvoice"""
        url = f"{self.__base_url}/tg-invoices/{invoice_id}"
        response = await self._request(self.__payment_name, self.__delete_method, url, headers=self.__headers)
        return True if response['success'] else False

    async def get_challenge(self, challenge_id: str, user_id: str) -> str:
        """Get challenge amount by user id.
        
        :param challenge_id: ID of the challenge.
        :param user_id: Telegram ID of the user.

        Docs: https://pay.xrocket.tg/api/#/challenges/ChallengesController_getTradeAmount
        
        :return: Amount in USD."""
        url = f"{self.__base_url}/challenges/{challenge_id}/users/{user_id}"
        response = await self._request(self.__payment_name, self.__get_method, url, headers=self.__headers)
        return response['data']['amountUsd']
    
    async def get_available_currencies(self) -> List[Currency]:
        """Returns available currencies.
        
        Docs: https://pay.xrocket.tg/api/#/currencies/CurrenciesController_getCoins"""
        url = f"{self.__base_url}/currencies/available"
        response = await self._request(self.__payment_name, self.__get_method, url, headers=self.__headers)
        return [Currency(**currency) for currency in response['data']['results']]
    
    async def create_subscription(self, interval: str, amount: float, status: str, referralPercent: int, currency: str, name: Optional[str] = None,
                                  description: Optional[str] = None, tgResource: Optional[str] = None, returnUrl: Optional[str] = None) -> Subscription:
        """Create subscription.
        
        :param interval: Interval for subscription (AsyncPayments.xrocket.models.SubscriptionsIntervals).
        :param amount: Cost subscription for current interval in currency.
        :param status: Status for subscription (AsyncPayments.xrocket.models.SubscriptionsStatuses).
        :param referralPercent: Subscription referral percent.
        :param currency: Subscription currency.
        :param name: Subscription name, view in bot.
        :param description: Subscription description, view in bot.
        :param tgResource: Subscription TG resource.
        :param returlUrl: Return link after payment.
        
        Docs: https://pay.xrocket.tg/api/#/subscriptions/SubscriptionsController_createSubscription"""
        params = {
            "name": name,
            "description": description,
            "currency": currency,
            "interval": [
                {
                "interval": interval,
                "amount": amount,
                "status": status
                }
            ],
            "tgResource": tgResource,
            "referralPercent": referralPercent,
            "returnUrl": returnUrl
        }
        self._delete_empty_fields(params)
        url = f"{self.__base_url}/subscriptions"
        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers)
        return Subscription(**response['data'])

    async def get_list_subscriptions(self, limit: Optional[int] = 100, offset: Optional[int] = 0) -> SubscriptionsList:
        """Get list of subscription.
        
        :param limit: Limit of subscriptions.
        :param offset: Offset.

        Docs: https://pay.xrocket.tg/api/#/subscriptions/SubscriptionsController_getSubscriptions"""
        params = {
            "limit": limit,
            "offset": offset,
        }
        url = f"{self.__base_url}/subscriptions?{urlencode(params)}"
        response = await self._request(self.__payment_name, self.__get_method, url, headers=self.__headers)
        return SubscriptionsList(**response['data'])

    async def get_subscription_info(self, subscription_id: int) -> Subscription:
        """Get subscription info.
        
        :param subscription_id: ID of the subscription.

        Docs: https://pay.xrocket.tg/api/#/subscriptions/SubscriptionsController_getSubscription"""
        url = f"{self.__base_url}/subscriptions/{subscription_id}"
        response = await self._request(self.__payment_name, self.__get_method, url, headers=self.__headers)
        return Subscription(**response['data'])
    
    async def delete_subscription(self, subscription_id: int) -> bool:
        """Delete subscription.
        
        :param subscription_id: ID of the subscription.

        Docs: https://pay.xrocket.tg/api/#/subscriptions/SubscriptionsController_deleteSubscription"""
        url = f"{self.__base_url}/subscriptions/{subscription_id}"
        response = await self._request(self.__payment_name, self.__delete_method, url, headers=self.__headers)
        return True if response['success'] else False
    
    async def check_subscription(self, subscription_id: int, user_id: int) -> SubscriptionCheck:
        """Delete subscription.
        
        :param subscription_id: ID of the subscription.
        :param user_id: ID of the user.

        Docs: https://pay.xrocket.tg/api/#/subscriptions/SubscriptionsController_checkSubscription"""
        params = {
            "userId": user_id,
        }
        url = f"{self.__base_url}/subscriptions/check/{subscription_id}"
        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, json=params)
        return SubscriptionCheck(**response['data'])
    
    async def get_subscription_interval_info(self, subscription_id: int, interval_code: str) -> Subscriptions.Interval:
        """Get subscription interval info.
        
        :param subscription_id: ID of the subscription.
        :param interval_code: Code of the interval.

        Docs: https://pay.xrocket.tg/api/#/subscriptions/SubscriptionsController_getSubscriptionInterval"""
        url = f"{self.__base_url}/subscriptions/{subscription_id}/interval/{interval_code}"
        response = await self._request(self.__payment_name, self.__get_method, url, headers=self.__headers)
        return Subscriptions.Interval(**response['data'])

    async def edit_subscription_interval(self, subscription_id: int, interval_code: str, status: str) -> Subscriptions.Interval:
        """Edit subscription interval.
        
        :param subscription_id: ID of the subscription.
        :param interval_code: Code of the interval.
        :param status: Status for subscription (AsyncPayments.xrocket.models.SubscriptionsStatuses).

        Docs: https://pay.xrocket.tg/api/#/subscriptions/SubscriptionsController_editSubscriptionInterval"""
        params = {
            "status": status
        }
        url = f"{self.__base_url}/subscriptions/{subscription_id}/interval/{interval_code}"
        response = await self._request(self.__payment_name, self.__put_method, url, headers=self.__headers, json=params)
        return Subscriptions.Interval(**response['data'])
    
    async def delete_subscription_interval(self, subscription_id: int, interval_code: str) -> Subscriptions.Interval:
        """Delete subscription interval.
        
        :param subscription_id: ID of the subscription.
        :param interval_code: Code of the interval.

        Docs: https://pay.xrocket.tg/api/#/subscriptions/SubscriptionsController_deleteSubscriptionInterval"""
        url = f"{self.__base_url}/subscriptions/{subscription_id}/interval/{interval_code}"
        response = await self._request(self.__payment_name, self.__delete_method, url, headers=self.__headers)
        return Subscriptions.Interval(**response['data'])
    
    async def create_subscription_interval(self, subscription_id: int, interval: str, amount: float, status: str) -> Subscriptions.Interval:
        """Create subscription interval.
        
        :param subscription_id: ID of the subscription.
        :param interval: Interval for subscription (AsyncPayments.xrocket.models.SubscriptionsIntervals).
        :param amount: Cost subscription for current interval in currency.
        :param status: Status for subscription (AsyncPayments.xrocket.models.SubscriptionsStatuses).

        Docs: https://pay.xrocket.tg/api/#/subscriptions/SubscriptionsController_creteSubscriptionInterval"""
        params = {
            "interval": interval,
            "amount": amount,
            "status": status,
        }
        url = f"{self.__base_url}/subscriptions/{subscription_id}"
        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers)
        return Subscriptions.Interval(**response['data'])


from pydantic import BaseModel
from typing import Optional, Union


class Balance(BaseModel):
    balance: Optional[int] = None
    ref_balance: Optional[Union[str, float, int]] = None
    
    
class Transaction(BaseModel):
    transaction: Optional[Union[str, int]] = None
    email: Optional[str] = None
    amount: Optional[Union[str, int]] = None
    currency: Optional[str] = None
    currency_amount: Optional[Union[str, int]] = None
    comission_percent: Optional[Union[str, int]] = None
    comission_fixed: Optional[str] = None
    amount_profit: Optional[Union[str, int]] = None
    method: Optional[str] = None
    payment_id: Optional[Union[int, str]] = None
    description: Optional[str] = None
    date: Optional[str] = None
    pay_date: Optional[str] = None
    transaction_status: Optional[Union[str, int]] = None
    custom_fields: Optional[Union[str, int, dict]] = None
    webhook_status: Optional[Union[str, int]] = None
    webhook_amount: Optional[Union[str, int]] = None
    
    
class Payout(BaseModel):
    payout: Optional[int] = None
    method: Optional[str] = None
    reciever: Optional[str] = None
    type: Optional[str] = None
    amount: Optional[int] = None
    comission_percent: Optional[int] = None
    comission_fixed: Optional[str] = None
    amount_profit: Optional[int] = None
    date_create: Optional[str] = None
    date_pay: Optional[str] = None
    status: Optional[int] = None
    

class PayoutOnCreate(BaseModel):
    payout_id: Optional[int] = None
    method: Optional[str] = None
    reciever: Optional[str] = None
    amount: Optional[int] = None
    comission_percent: Optional[int] = None
    comission_fixed: Optional[str] = None
    amount_profit: Optional[int] = None
    date: Optional[str] = None
    payout_status_code: Optional[int] = None
    payout_status_text: Optional[str] = None
    
    
class CreatePayout(BaseModel):
    remain_balance: Optional[Union[str, int, float]] = None
    payout: Optional[PayoutOnCreate] = None
    


from typing import List
from ..requests import RequestsClient
from ..exceptions import InvalidRequest, InvalidGrant, EmptyToken, UnauthorizedClient

class Authorize(RequestsClient):
    def __init__(self, client_id: str, redirect_uri: str, client_secret: str, scope: List[str]):
        super().__init__()
        self.client_id = client_id
        self.redirect_uri = redirect_uri
        self.client_secret = client_secret
        self.scope = scope
    
    async def authorize(self):
        url = "https://yoomoney.ru/oauth/authorize?client_id={client_id}&response_type=code" \
              "&redirect_uri={redirect_uri}&scope={scope}".format(client_id=self.client_id,
                                                                  redirect_uri=self.redirect_uri,
                                                                  scope='%20'.join([str(elem) for elem in self.scope]),
                                                                  )

        headers = {
            'Content-Type': 'application/x-www-form-urlencoded'
        }
        response = await self._request_for_authorize_yoomoney("POST", url, headers=headers)
        if response.status == 200:
            print("Visit this website and confirm the application authorization request:")
            print(response.url)

        code = str(input("Enter redirected url (https://yourredirect_uri?code=XXXXXXXXXXXXX) or just code: "))
        try:
            code = code[code.index("code=") + 5:].replace(" ","")
        except:
            pass

        url = "https://yoomoney.ru/oauth/token?code={code}&client_id={client_id}&" \
              "grant_type=authorization_code&redirect_uri={redirect_uri}&client_secret={client_secret}".format(code=str(code), client_id=self.client_id, 
                                                                                                               redirect_uri=self.redirect_uri, 
                                                                                                               client_secret=self.client_secret
                                                                                                               )
        response = await self._request("yoomoney", "POST", url, headers=headers)
        
        if "error" in response:
            error = response["error"]
            if error == "invalid_request":
                raise InvalidRequest("Required query parameters are missing or have incorrect or invalid values.")
            elif error == "unauthorized_client":
                raise UnauthorizedClient("Invalid parameter value 'client_id' or 'client_secret', or the application" \
              " does not have the right to request authorization (for example, YooMoney blocked it 'client_id').")
            elif error == "invalid_grant":
                raise InvalidGrant("In issue 'access_token' denied. YuMoney did not issue a temporary token, " \
              "the token is expired, or this temporary token has already been issued " \
              "'access_token' (repeated request for an authorization token with the same temporary token).")

        if response['access_token'] == "":
            raise EmptyToken("Response token is empty. Repeated request for an authorization token.")

        print("Your access token:")
        print(response['access_token'])


from pydantic import BaseModel
from typing import List, Optional


class MeInfo(BaseModel):
    app_id: Optional[int] = None
    name: Optional[str] = None
    payment_processing_bot_username: Optional[str] = None

class Invoice(BaseModel):
    invoice_id: Optional[int] = None
    hash: Optional[str] = None
    currency_type: Optional[str] = None
    asset: Optional[str] = None
    fiat: Optional[str] = None
    amount: Optional[str] = None
    paid_asset: Optional[str] = None
    paid_amount: Optional[str] = None
    paid_fiat_rate: Optional[str] = None
    accepted_assets: Optional[List[str]] = None
    fee_asset: Optional[str] = None
    fee_amount: Optional[float] = None
    fee: Optional[str] = None
    bot_invoice_url: Optional[str] = None
    pay_url: Optional[str] = None
    description: Optional[str] = None
    status: Optional[str] = None
    created_at: Optional[str] = None
    paid_usd_rate: Optional[str] = None
    usd_rate: Optional[str] = None
    allow_comments: Optional[bool] = None
    allow_anonymous: Optional[bool] = None
    expiration_date: Optional[str] = None
    paid_at: Optional[str] = None
    paid_anonymously: Optional[bool] = None
    comment: Optional[str] = None
    hidden_message: Optional[str] = None
    payload: Optional[str] = None
    paid_btn_name: Optional[str] = None
    paid_btn_url: Optional[str] = None

class Check(BaseModel):
    check_id: Optional[int] = None
    hash: Optional[str] = None
    asset: Optional[str] = None
    amount: Optional[float] = None
    bot_check_url: Optional[str] = None
    status: Optional[str] = None
    created_at: Optional[str] = None
    activated_at: Optional[str] = None

class Transfer(BaseModel):
    transfer_id: Optional[int] = None
    user_id: Optional[int] = None
    asset: Optional[str] = None
    amount: Optional[float] = None
    status: Optional[str] = None
    completed_at: Optional[str] = None
    comment: Optional[str] = None

class Balance(BaseModel):
    currency_code: Optional[str] = None
    available: Optional[float] = None
    onhold: Optional[float] = None

class ExchangeRate(BaseModel):
    is_valid: Optional[bool] = None
    is_crypto: Optional[bool] = None
    is_fiat: Optional[bool] = None
    source: Optional[str] = None
    target: Optional[str] = None
    rate: Optional[float] = None

class Currency(BaseModel):
    is_blockchain: Optional[bool] = None
    is_stablecoin: Optional[bool] = None
    is_fiat: Optional[bool] = None
    name: Optional[str] = None
    code: Optional[str] = None
    url: Optional[str] = None
    decimals: Optional[int] = None



from ..requests import RequestsClient
from typing import Optional, Union, List
from .models import CreatePayment, CassaInfo, PayoffCreate, TickersRate, PayoffRequest, \
                    PaymentInfo, Balance, SwapPair, \
                    CreateSwap, SwapInfo, CreateTransfer, TransferInfo, Stats
import json
import hashlib


class AsyncCrystalPay(RequestsClient):
    API_HOST: str = "https://crystalpay.io"

    def __init__(self, login: str, secret: str, salt: str) -> None:
        """
        Initialize CrystalPay API client
        :param login: Your Login
        :param secret: Your Secret
        :param salt: Your Salt
        """
        super().__init__()
        self.__login = login
        self.__secret = secret
        self.__salt = salt
        self.__headers = {
            'Content-Type': 'application/json',
        }
        self.__base_url = "https://api.crystalpay.io/v3"
        self.__post_method = "POST"
        self.__payment_name = "crystalPay"
        self.check_values()

    def check_values(self):
        if not self.__login or not self.__secret or not self.__salt:
            raise ValueError('No Secret, Login or Salt specified')

    async def get_cassa_info(self, hide_empty: Optional[bool]= False) -> CassaInfo:
        """Get cash info.

        Docs: https://docs.crystalpay.io/metody-api/me-kassa/poluchenie-informacii-o-kasse"""

        url = f'{self.__base_url}/me/info/'

        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "hide_empty": hide_empty,
        }
        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))

        return CassaInfo(**response)

    async def get_balance_list(self, hide_empty: Optional[bool] = False) -> dict:
        """Get balances list.

        Docs: https://docs.crystalpay.io/metody-api/balance-balansy/poluchenie-spiska-balansov
        
        :param hide_empty: Optional. Hide empty balances. Defaults to False"""

        url = f'{self.__base_url}/balance/list/'

        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "hide_empty": hide_empty,
        }
        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))
        if response['items']:
            return response['items']
        else:
            return []
    
    async def get_balance(self, method: str) -> Balance:
        """Get balance of the method.

        Docs: https://docs.crystalpay.io/metody-api/balance-balansy/poluchenie-balansa
        
        :param method: Internal name of the account/method."""

        url = f'{self.__base_url}/balance/get/'

        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "method": method,
        }
        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))
        return Balance(**response)
    
    async def get_payment_methods(self, compact: Optional[bool] = False) -> dict:
        """Get list of a methods.

        Docs: https://docs.crystalpay.io/metody-api/method-metody/poluchenie-spiska-metodov
        
        :param compact: Displaying only basic information."""

        url = f'{self.__base_url}/method/list/'

        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "compact": compact,
        }
        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))
        
        return response['items']
    
    
    async def get_payment_method(self, method: str) -> dict:
        """Get info about a method.

        Docs: https://docs.crystalpay.io/metody-api/method-metody/poluchenie-metoda
        
        :param compact: The internal name of the method."""

        url = f'{self.__base_url}/method/get/'

        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "method": method,
        }
        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))

        return response
        

    async def edit_payment_method(self, method: str, extra_commission_percent: Union[int, float], enabled: bool) -> bool:
        """Edit payment method

        Dosc: https://docs.crystalpay.io/api/metody-oplaty/izmenenie-nastroek-metoda-oplaty

        :param method: Payment method, for example: LZTMARKET, BITCOIN
        :param extra_commission_percent: Additional cash desk commission for payment method, in percent
        :param enabled: Enable/Disable payment method

        :return: True if successful, else exception BadRequest
        """

        url = f'{self.__base_url}/method/edit/'

        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "method": method,
            "extra_commission_percent": extra_commission_percent,
            "enabled": enabled,
        }

        await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))

        return True

    async def create_payment(
            self, amount: Union[int, float], amount_currency: Optional[str] = "RUB", required_methods: Optional[str] = None,
            _type: Optional[str] = "purchase", description: Optional[str] = None, redirect_url: Optional[str] = None,
            callback_url: Optional[str] = None, extra: Optional[str] = None, payer_details: Optional[str] = None,
            lifetime: Optional[int] = 60) -> CreatePayment:
        """Generate payment url.

        Docs: https://docs.crystalpay.io/metody-api/invoice-platezhi/sozdanie-invoisa

        :param amount: Order amount.
        :param amount_currency: Currency. Default to 'RUB', for example: USD, BTC, ETH
        :param required_methods: Pre-selected payment method, for example: LZTMARKET, BITCOIN
        :param _type: Invoice type. possible options: purchase, topup
        :param description: The description or purpose of the payment is displayed to the payer on the payment page, for example: Account purchase #12345678
        :param redirect_url: Redirect link after payment
        :param callback_url: Link for HTTP Callback notification after successful payment
        :param extra: Any internal data, for example: Payment ID in your system
        :param payer_details: Payer email
        :param lifetime: Invoice lifetime in minutes, maximum - 4320. Default to 60"""

        url = f'{self.__base_url}/invoice/create/'

        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "amount": amount,
            "amount_currency": amount_currency,
            "required_methods": required_methods,
            "type": _type,
            "description": description,
            "redirect_url": redirect_url,
            "callback_url": callback_url,
            "extra": extra,
            "payer_details": payer_details,
            "lifetime": lifetime,
        }
        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))

        return CreatePayment(**response)

    async def get_payment_info(self, invoice_id: str) -> PaymentInfo:
        """Get info about payment.

        Docs: https://docs.crystalpay.io/metody-api/invoice-platezhi/poluchenie-informacii-ob-invoise

        :param invoice_id: Invoice ID"""

        url = f'{self.__base_url}/invoice/info/'

        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "id": invoice_id,
        }
        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))

        return PaymentInfo(**response)

    async def create_payoff(self, amount: Union[int, float], method: str, wallet: str, subtract_from: str,
                            amount_currency: Optional[str] = None, callback_url: Optional[str] = None,
                            extra: Optional[str] = None, wallet_extra: Optional[str] = None) -> PayoffCreate:
        """Create payoff request.

        Docs: https://docs.crystalpay.io/metody-api/payoff-vyvody/sozdanie-vyvoda

        :param amount: Payoff amount, for example: 10, 0.0015
        :param amount_currency: The currency of the amount is automatically converted into the currency of the withdrawal method, for example: RUB, USD, BTC
        :param method: Payoff method, for example: LZTMARKET, BITCOIN
        :param wallet: Recipient's wallet details
        :param subtract_from: Where to write off the commission amount, possible options: balance, amount
        :param callback_url: Link for HTTP Callback notification after output is complete
        :param extra: Any internal data, for example: Payment ID in your system
        :param wallet_extra: Optional. Additional information about the recipient's details.

        More about subtract_from:

        amount - The commission will be deducted from the withdrawal amount. The amount will be credited to your wallet.
        balance - The commission will be deducted from the balance. The exact amount will be sent to your wallet."""

        url = f'{self.__base_url}/payoff/create/'

        signature = hashlib.sha1(str.encode(f"{amount}:{method}:{wallet}:{self.__salt}")).hexdigest()

        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "signature": signature,
            "amount": amount,
            "amount_currency": amount_currency,
            "method": method,
            "wallet": wallet,
            "wallet_extra": wallet_extra,
            "subtract_from": subtract_from,
            "callback_url": callback_url,
            "extra": extra,
        }
        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))

        return PayoffCreate(**response)

    async def submit_payoff(self, payoff_id: str) -> PayoffRequest:
        """Submit payoff request

        Docs: https://docs.crystalpay.io/metody-api/payoff-vyvody/podtverzhdenie-vyvoda

        :param payoff_id: Payoff ID"""

        signature = hashlib.sha1(str.encode(f"{payoff_id}:{self.__salt}")).hexdigest()

        url = f"{self.__base_url}/payoff/submit/"

        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "signature": signature,
            "id": payoff_id,
        }

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))

        return PayoffRequest(**response)


    async def cancel_payoff(self, payoff_id: str) -> PayoffRequest:
        """Cancel payoff request

        Docs: https://docs.crystalpay.io/metody-api/payoff-vyvody/otmena-vyvoda

        :param payoff_id: Payoff ID"""

        signature = hashlib.sha1(str.encode(f"{payoff_id}:{self.__salt}")).hexdigest()

        url = f"{self.__base_url}/payoff/cancel/"

        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "signature": signature,
            "id": payoff_id,
        }

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))

        return PayoffRequest(**response)

    async def get_payoff(self, payoff_id: str) -> PayoffRequest:
        """Get info about payoff request

        Docs: https://docs.crystalpay.io/metody-api/payoff-vyvody/poluchenie-informacii-o-vyvode

        :param payoff_id: Payoff ID"""

        url = f"{self.__base_url}/payoff/info/"

        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "id": payoff_id,
        }

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))

        return PayoffRequest(**response)

    async def get_tickers_list(self) -> list:
        """Get a list of available currencies

        Dosc: https://docs.crystalpay.io/api/valyuty/poluchenie-spiska-dostupnykh-valyut"""

        url = f"{self.__base_url}/ticker/list/"

        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
        }

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))

        return list(response['tickers'])

    async def get_tickers_rate(self, tickers: list, base_currency: Optional[str] = "RUB") -> TickersRate:
        """Get the exchange rate against 

        Docs: https://docs.crystalpay.io/api/valyuty/poluchenie-kursa-valyut

        :param tickers: Array of currencies, for example: [“BTC”, “LTC”]
        :param base_currency: Base currency, RUB by default."""

        url = f"{self.__base_url}/ticker/get/"

        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "tickers": tickers,
            "base_currency": base_currency,
        }

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))

        return TickersRate(**response)

    async def get_list_swap_pairs(self, page: Optional[int] = 1, items: Optional[int] = 20, source: Optional[str] = None,
                                      target: Optional[str] = None) -> List[SwapPair]:
        """Getting a list of swap pairs.

        Docs: https://docs.crystalpay.io/metody-api/swap-obmeny/obmennye-pary/poluchenie-spiska-obmennykh-par

        :param page: Optional. Page Number. Default is 1.
        :param items: Optional. Number of items per page. Default is 20.
        :param source: Optional. The original currency, deducted during the exchange. Filter by source.
        :param target: Optional. The currency received is credited during the exchange. Filter by target.
        """

        url = f"{self.__base_url}/swap/pair/list/"

        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "page": page,
            "items": items,
            "source": source,
            "target": target,
        }
        self._delete_empty_fields(params)

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))

        return [SwapPair(**swapPair) for swapPair in response['items'].values()]
    
    async def get_swap_pair(self, pair_id: int) -> SwapPair:
        """Get a swap pair.

        Docs: https://docs.crystalpay.io/metody-api/swap-obmeny/obmennye-pary/poluchenie-obmennoi-pary

        :param pair_id: ID of the swap pair.
        """

        url = f"{self.__base_url}/swap/pair/get/"

        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "pair_id": pair_id,
        }

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))

        return SwapPair(**response)
    
    async def create_swap(self, pair_id: int, amount: int, amount_type: str) -> CreateSwap:
        """Create swap.

        Docs: https://docs.crystalpay.io/metody-api/swap-obmeny/sozdanie-obmena

        :param pair_id: ID of the swap pair.
        :param amount: Amount of the swap.
        :param amount_type: Type of the amount. source - The amount specified in the original currency. target - The amount specified in the received currency.
        """
        
        url = f"{self.__base_url}/swap/create/"
        signature = hashlib.sha1(str.encode(f"{amount}:{pair_id}:{self.__salt}")).hexdigest()

        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "pair_id": pair_id,
            "amount": amount,
            "amount_type": amount_type,
            "signature": signature,
        }

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))

        return CreateSwap(**response)
    
    async def swap_submit(self, swap_id: str) -> SwapInfo:
        """Submit swap request.

        Docs: https://docs.crystalpay.io/metody-api/swap-obmeny/podtverzhdenie-obmena

        :param swap_id: ID of the swap.
        """
        
        url = f"{self.__base_url}/swap/submit/"
        signature = hashlib.sha1(str.encode(f"{swap_id}:{self.__salt}")).hexdigest()

        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "id": swap_id,
            "signature": signature,
        }

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))

        return SwapInfo(**response)
    
    async def swap_cancel(self, swap_id: str) -> SwapInfo:
        """Cancel swap request.

        Docs: https://docs.crystalpay.io/metody-api/swap-obmeny/otmena-obmena

        :param swap_id: ID of the swap.
        """
        
        url = f"{self.__base_url}/swap/cancel/"
        signature = hashlib.sha1(str.encode(f"{swap_id}:{self.__salt}")).hexdigest()

        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "id": swap_id,
            "signature": signature,
        }

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))

        return SwapInfo(**response)
    
    async def get_swap_info(self, swap_id: str) -> SwapInfo:
        """Get swap info.

        Docs: https://docs.crystalpay.io/metody-api/swap-obmeny/poluchenie-informacii-ob-obmene

        :param swap_id: ID of the swap.
        """
        
        url = f"{self.__base_url}/swap/info/"
        

        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "id": swap_id,
        }

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))

        return SwapInfo(**response)
    
    async def create_transfer(self, method: str, amount: str, receiver: str, 
                              description: Optional[str] = None) -> CreateTransfer:
        """Create a transfer.

        Docs: https://docs.crystalpay.io/metody-api/transfer-perevody/sozdanie-perevoda

        :param method: The internal name of the method.
        :param amount: Amount of the transfer.
        :param receiver: The recipient's cashier's login.
        :param description: Optional. Description of the transfer.
        """
        
        url = f"{self.__base_url}/transfer/create/"
        signature = hashlib.sha1(str.encode(f"{amount}:{method}:{receiver}:{self.__salt}")).hexdigest()
        
        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "method": method,
            "amount": amount,
            "receiver": receiver,
            "signature": signature,
            "description": description,
        }

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))

        return CreateTransfer(**response)
    
    async def submit_transfer(self, transfer_id: str) -> TransferInfo:
        """Submit the transfer.

        Docs: https://docs.crystalpay.io/metody-api/transfer-perevody/podtverzhdenie-perevoda

        :param transfer_id: ID of the transfer.
        """
        
        url = f"{self.__base_url}/transfer/submit/"
        signature = hashlib.sha1(str.encode(f"{transfer_id}:{self.__salt}")).hexdigest()
        
        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "id": transfer_id,
            "signature": signature,
        }

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))

        return TransferInfo(**response)
    
    async def cancel_transfer(self, transfer_id: str) -> TransferInfo:
        """Cancel the transfer.

        Docs: https://docs.crystalpay.io/metody-api/transfer-perevody/otmena-perevoda

        :param transfer_id: ID of the transfer.
        """
        
        url = f"{self.__base_url}/transfer/submit/"
        signature = hashlib.sha1(str.encode(f"{transfer_id}:{self.__salt}")).hexdigest()
        
        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "id": transfer_id,
            "signature": signature,
        }

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))

        return TransferInfo(**response)

    async def get_transfer_info(self, transfer_id: str) -> TransferInfo:
        """Get info about the transfer.

        Docs: https://docs.crystalpay.io/metody-api/transfer-perevody/poluchenie-informacii-o-perevode

        :param transfer_id: ID of the transfer.
        """
        
        url = f"{self.__base_url}/transfer/info/"
        
        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "id": transfer_id,
        }

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))

        return TransferInfo(**response)
    
    async def get_history_payments(self, page: Optional[int] = 1, items: Optional[int] = 20, period: Optional[int] = 1, 
                                   export_csv: Optional[bool] = False) -> List[PaymentInfo]:
        """Get payments history.

        Docs: https://docs.crystalpay.io/metody-api/report-otchyoty-i-statistika/invoice-platezhi/poluchenie-istorii

        :param page: Optional. Page number. Default is 1.
        :param items: Optional. Number of items per page. Default is 20.
        :param period: Optional. The period from the current date, in days. Default is 1.
        :param export_csv: Optional. Export as a csv format table. Default is False.
        """
        
        url = f"{self.__base_url}/report/invoice/history/"
        
        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "page": page,
            "items": items,
            "period": period,
            "export_csv": export_csv,
        }

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))

        if response["items"]:
            return [PaymentInfo(**payment) for payment in response['items']]
        else:
            return []
    
    async def get_stats_payments(self, period: Optional[int] = 1, 
                                 export_pdf: Optional[bool] = False) -> Stats:
        """Get payments stats.

        Docs: https://docs.crystalpay.io/metody-api/report-otchyoty-i-statistika/invoice-platezhi/poluchenie-statistiki

        :param period: Optional. The period from the current date, in days. Default is 1.
        :param export_pdf: Optional. Export as a pdf format. Default is False.
        """
        
        url = f"{self.__base_url}/report/invoice/summary/"
        
        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "period": period,
            "export_pdf": export_pdf,
        }

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))

        return Stats(**response)
    
    async def get_history_payoffs(self, page: Optional[int] = 1, items: Optional[int] = 20, period: Optional[int] = 1, 
                                   export_csv: Optional[bool] = False) -> List[PayoffRequest]:
        """Get payoffs history.

        Docs: https://docs.crystalpay.io/metody-api/report-otchyoty-i-statistika/payoff-vyvody/poluchenie-istorii

        :param page: Optional. Page number. Default is 1.
        :param items: Optional. Number of items per page. Default is 20.
        :param period: Optional. The period from the current date, in days. Default is 1.
        :param export_csv: Optional. Export as a csv format table. Default is False.
        """
        
        url = f"{self.__base_url}/report/payoff/history/"
        
        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "page": page,
            "items": items,
            "period": period,
            "export_csv": export_csv,
        }

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))

        if response['items']:
            return [PayoffRequest(**payment) for payment in response['items']]
        else:
            return []
    
    async def get_stats_payoffs(self, period: Optional[int] = 1, 
                                 export_pdf: Optional[bool] = False) -> Stats:
        """Get payoff stats.

        Docs: https://docs.crystalpay.io/metody-api/report-otchyoty-i-statistika/payoff-vyvody/poluchenie-statistiki

        :param period: Optional. The period from the current date, in days. Default is 1.
        :param export_pdf: Optional. Export as a pdf format. Default is False.
        """
        
        url = f"{self.__base_url}/report/payoff/summary/"
        
        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "period": period,
            "export_pdf": export_pdf,
        }

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))

        return Stats(**response)
    
    
    async def get_history_swaps(self, page: Optional[int] = 1, items: Optional[int] = 20, period: Optional[int] = 1, 
                                   export_csv: Optional[bool] = False) -> List[SwapInfo]:
        """Get swaps history.

        Docs: https://docs.crystalpay.io/metody-api/report-otchyoty-i-statistika/swap-obmeny/poluchenie-istorii

        :param page: Optional. Page number. Default is 1.
        :param items: Optional. Number of items per page. Default is 20.
        :param period: Optional. The period from the current date, in days. Default is 1.
        :param export_csv: Optional. Export as a csv format table. Default is False.
        """
        
        url = f"{self.__base_url}/report/swap/history/"
        
        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "page": page,
            "items": items,
            "period": period,
            "export_csv": export_csv,
        }

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))
        
        if response["items"]:
            return [SwapInfo(**payment) for payment in response['items']]
        else:
            return []
    
    async def get_history_transfers(self, page: Optional[int] = 1, items: Optional[int] = 20, period: Optional[int] = 1, 
                                   export_csv: Optional[bool] = False) -> List[TransferInfo]:
        """Get swaps history.

        Docs: https://docs.crystalpay.io/metody-api/report-otchyoty-i-statistika/transfer-perevody/poluchenie-istorii

        :param page: Optional. Page number. Default is 1.
        :param items: Optional. Number of items per page. Default is 20.
        :param period: Optional. The period from the current date, in days. Default is 1.
        :param export_csv: Optional. Export as a csv format table. Default is False.
        """
        
        url = f"{self.__base_url}/report/transfer/history/"
        
        params = {
            "auth_login": self.__login,
            "auth_secret": self.__secret,
            "page": page,
            "items": items,
            "period": period,
            "export_csv": export_csv,
        }

        response = await self._request(self.__payment_name, self.__post_method, url, headers=self.__headers, data=json.dumps(params))
        
        if response["items"]:
            return [TransferInfo(**payment) for payment in response['items']]
        else:
            return []
    


from pydantic import BaseModel, Field
from typing import Optional, Union


class OrderStatuses:
    PENDING: str = "PENDING"
    CANCELED: str = "CANCELED"
    CONFIRMED: str = "CONFIRMED"
    CHARGEBACKED: str = "CHARGEBACKED"


class Order(BaseModel):
    paymentMethod: Optional[str] = None
    transactionId: Optional[str] = None
    redirect: Optional[str] = None
    return_: Optional[str] = Field(alias="return", default=None)
    paymentDetails: Optional[Union[str, dict]] = None
    status: Optional[str] = None
    expiresIn: Optional[str] = None
    merchantId: Optional[str] = None
    usdtRate: Optional[Union[int, float,]] = None
    
    
class OrderInfo(BaseModel):
    id: Optional[str] = None
    status: Optional[str] = None
    paymentDetails: Optional[dict] = None
    merchantName: Optional[str] = None
    merchantId: Optional[str] = None
    comission: Optional[Union[int, float]] = None
    paymentMethod: Optional[str] = None
    expiresIn: Optional[str] = None
    return_: Optional[str] = Field(alias="return", default=None)
    comissionUsdt: Optional[Union[int, float]] = None
    amountUsdt: Optional[Union[int, float]] = None
    qr: Optional[str] = None
    payformSuccessUrl: Optional[str] = None
    payload: Optional[str] = None
    comissionType: Optional[int] = None
    externalId: Optional[str] = None
    description: Optional[str] = None
    
    
class ExchangeRate(BaseModel):
    paymentMethod: Optional[int] = None
    currencyFrom: Optional[str] = None
    currencyTo: Optional[str] = None
    rate: Optional[float] = None
    updatedAt: Optional[str] = None


from pydantic import BaseModel, Field
from typing import Optional, Union, List


class Balances(BaseModel):
    merchant: Optional[list] = None
    user: Optional[list] = None


class Balance(BaseModel):
    uuid: Optional[str] = None
    balance: Optional[str] = None
    currency_code: Optional[str] = None
    balance_usd: Optional[str] = None


class CreatePayment(BaseModel):
    uuid: Optional[str] = None
    order_id: Optional[str] = None
    amount: Optional[str] = None
    payment_amount: Optional[Union[str, int]] = None
    payment_amount_usd: Optional[Union[str, int]] = None
    payer_amount: Optional[Union[str, int]] = None
    payer_amount_exchange_rate: Optional[Union[str, int]] = None
    discount_percent: Optional[Union[str, int]] = None
    discount: Optional[str] = None
    payer_currency: Optional[Union[str, int]] = None
    currency: Optional[str] = None
    comments: Optional[Union[str, dict, list]] = None
    merchant_amount: Optional[Union[str, int]] = None
    network: Optional[str] = None
    address: Optional[str] = None
    from_: Optional[str] = Field(alias="from", default=None)
    txid: Optional[str] = None
    payment_status: Optional[str] = None
    url: Optional[str] = None
    expired_at: Optional[int] = None
    status: Optional[str] = None
    is_final: Optional[bool] = None
    aditional_data: Optional[str] = None
    created_at: Optional[str] = None
    updated_at: Optional[str] = None
    commission: Optional[Union[str, int]] = None
    address_qr_code: Optional[str] = None
    mercuryo_payment_link: Optional[str] = None


class GenerateStaticWallet(BaseModel):
    wallet_uuid: Optional[str] = None
    uuid: Optional[str] = None
    address: Optional[str] = None
    network: Optional[str] = None
    currency: Optional[str] = None
    url: Optional[str] = None
    

class GenerateQrCode(BaseModel):
    image: Optional[str] = None


class BlockStaticWallet(BaseModel):
    uuid: Optional[str] = None
    status: Optional[str] = None


class RefundPaymentsOnBlockedAddress(BaseModel):
    commission: Optional[str] = None
    amount: Optional[str] = None


class PaymentInfo(BaseModel):
    uuid: Optional[str] = None
    order_id: Optional[str] = None
    amount: Optional[str] = None
    payment_amount: Optional[str] = None
    payer_amount: Optional[str] = None
    discount_percent: Optional[Union[int, float, str]] = None
    discount: Optional[str] = None
    payer_currency: Optional[str] = None
    currency: Optional[str] = None
    merchant_amount: Optional[str] = None
    network: Optional[str] = None
    address: Optional[str] = None
    from_: Optional[str] = Field(alias="from", default=None)
    txid: Optional[str] = None
    payment_status: Optional[str] = None
    url: Optional[str] = None
    expired_at: Optional[int] = None
    status: Optional[str] = None
    is_final: Optional[bool] = None
    additional_data: Optional[str] = None
    created_at: Optional[str] = None
    updated_at: Optional[str] = None


class ServiceInfo:
    class ServiceLimit(BaseModel):
        min_amount: Optional[str] = None
        max_amount: Optional[str] = None


    class ServiceCommission(BaseModel):
        fee_amount: Optional[str] = None
        percent: Optional[str] = None


class ListOfServices(BaseModel):
    network: Optional[str] = None
    currency: Optional[str] = None
    is_available: Optional[bool] = None
    limit: Optional[ServiceInfo.ServiceLimit] = None
    commission: Optional[ServiceInfo.ServiceCommission] = None


class HistoryPaginate(BaseModel):
    count: Optional[int] = None
    hasPages: Optional[bool] = None
    nextCursor: Optional[str] = None
    previousCursor: Optional[str] = None
    perPage: Optional[int] = None


class PaymentHistory(BaseModel):
    items: Optional[List[PaymentInfo]] = None
    paginate: Optional[HistoryPaginate] = None


class PayoutHistoryItem(BaseModel):
    uuid: Optional[str] = None
    amount: Optional[str] = None
    currency: Optional[str] = None
    network: Optional[str] = None
    address: Optional[str] = None
    txid: Optional[str] = None
    status: Optional[str] = None
    is_final: Optional[bool] = None
    balance: Optional[Union[str, int]] = None
    created_at: Optional[str] = None
    updated_at: Optional[str] = None


class PayoutHistory(BaseModel):
    merchant_uuid: Optional[str] = None
    items: Optional[List[PayoutHistoryItem]] = None
    paginate: Optional[HistoryPaginate] = None


class Payout(BaseModel):
    uuid: Optional[str] = None
    amount: Optional[str] = None
    currency: Optional[str] = None
    network: Optional[str] = None
    address: Optional[str] = None
    txid: Optional[str] = None
    status: Optional[str] = None
    is_final: Optional[bool] = None
    balance: Optional[Union[str, int]] = None
    payer_currency: Optional[str] = None
    payer_amount: Optional[Union[str, int]] = None


class ListOfServicesPayout(BaseModel):
    network: Optional[str] = None
    currency: Optional[str] = None
    is_available: Optional[bool] = None
    limit: Optional[ServiceInfo.ServiceLimit] = None
    commission: Optional[ServiceInfo.ServiceCommission] = None


class TransferWallet(BaseModel):
    user_wallet_transaction_uuid: Optional[str] = None
    user_wallet_balance: Optional[str] = None
    merchant_transaction_uuid: Optional[str] = None
    merchant_balance: Optional[str] = None


class RecurringPayment(BaseModel):
    uuid: Optional[str] = None
    name: Optional[str] = None
    order_id: Optional[str] = None
    amount: Optional[str] = None
    currency: Optional[str] = None
    payer_currency: Optional[str] = None
    payer_amount_usd: Optional[str] = None
    payer_amount: Optional[str] = None
    url_callback: Optional[str] = None
    period: Optional[str] = None
    status: Optional[str] = None
    url: Optional[str] = None
    last_pay_off: Optional[str] = None
    additional_data: Optional[str] = None


class ListOfRecurringPayments(BaseModel):
    merchant_uuid: Optional[str] = None
    items: Optional[List[RecurringPayment]] = None
    paginate: Optional[HistoryPaginate] = None


class ExchangeRatesList(BaseModel):
    from_: Optional[str] = Field(alias="from", default=None)
    to: Optional[str] = None
    course: Optional[str] = None


class Discount(BaseModel):
    currency: Optional[str] = None
    network: Optional[str] = None
    discount: Optional[Union[str, int, float]] = None


class CurrenciesNames:
    USDC: str = "USDC"
    ETH: str = "ETH"
    USDT: str = "USDT"
    AVAX: str = "AVAX"
    BCH: str = "BCH"
    BNB: str = "BNB"
    CGPT: str = "CGPT"
    DAI: str = "DAI"
    BTC: str = "BTC"
    DASH: str = "DASH"
    DOGE: str = "DOGE"
    SHIB: str = "SHIB"
    MATIC: str = "MATIC"
    VERSE: str = "VERSE"
    LTC: str = "LTC"
    CRMS: str = "CRMS"
    SOL: str = "SOL"
    TON: str = "TON"
    HMSTR: str = "HMSTR"
    TRX: str = "TRX"
    XMR: str = "XMR"


class NetworkNames:
    ARBITRUM: str = "ARBITRUM"
    AVALANCHE: str = "AVALANCHE"
    BCH: str = "BCH"
    BSC: str = "BSC"
    BTC: str = "BTC"
    DASH: str = "DASH"
    DOGE: str = "DOGE"
    ETH: str = "ETH"
    LTC: str = "LTC"
    POLYGON: str = "POLYGON"
    SOL: str = "SOL"
    TON: str = "TON"
    TRON: str = "TRON"
    XMR: str = "XMR"


class InvoiceStatuses:
    PROCESS: str = "process"
    CHECK: str = "check"
    CONFIRM_CHECK: str = "confirm_check"
    PAID: str = "paid"
    PAID_OVER: str = "paid_over"
    FAIL: str = "fail"
    WRONG_AMOUNT: str = "wrong_amount"
    WRONG_AMOUNT_WAITING = "wrong_amount_waiting"
    CANCEL: str = "cancel"
    SYSTEM_FAIL: str = "system_fail"
    REFUND_PROCESS: str = "refund_process"
    REFUND_FAIL: str = "refund_fail"
    REFUND_PAID: str = "refund_paid"
    LOCKED: str = "locked"


class PayoutStatuses:
    PROCESS: str = "process"
    CHECK: str = "check"
    PAID: str = "paid"
    FAIL: str = "fail"
    CANCEL: str = "cancel"
    SYSTEM_FAIL: str = "system_fail"


from pydantic import BaseModel
from typing import Optional, List, Union


class Balance(BaseModel):
    currency: Optional[str] = None
    value: Optional[Union[int, str, float]] = None
    

class Order(BaseModel):
    merchant_order_id: Optional[str] = None
    fk_order_id: Optional[int] = None
    amount: Optional[Union[int, float]] = None
    currency: Optional[str] = None
    email: Optional[str] = None
    account: Optional[str] = None
    date: Optional[str] = None
    status: Optional[int] = None
    payer_account: Optional[str] = None
    
    
class Orders(BaseModel):
    pages: Optional[int] = None
    orders: Optional[List[Order]] = None
    
    
class CreateOrder(BaseModel):
    orderId: Optional[int] = None
    orderHash: Optional[str] = None
    location: Optional[str] = None
    
    
class Currency(BaseModel):
    id: Optional[int] = None
    name: Optional[str] = None
    currency: Optional[str] = None
    is_enabled: Optional[int] = None
    is_favorite: Optional[int] = None
    
    
class WithdrawalCurrency(BaseModel):
    id: Optional[int] = None
    name: Optional[str] = None
    min: Optional[int] = None
    max: Optional[int] = None
    currency: Optional[str] = None
    can_exchange: Optional[int] = None
    
    
class Store(BaseModel):
    id: Optional[int] = None
    name: Optional[str] = None
    url: Optional[str] = None
    activated: Optional[int] = None
    
    
class Withdrawal(BaseModel):
    id: Optional[int] = None
    amount: Optional[Union[int, float]] = None
    currency: Optional[str] = None
    ext_currency_id: Optional[int] = None
    account: Optional[str] = None
    date: Optional[str] = None
    status: Optional[int] = None
    
    
class Withdrawals(BaseModel):
    pages: Optional[int] = None
    withdrawals: Optional[List[Withdrawal]] = None
    
    
class CreateWithdrawal(BaseModel):
    id: Optional[int] = None
    


from ..requests import RequestsClient
from typing import Optional, Union, List
from .models import Balance, Transaction, Payout, CreatePayout, PayoutOnCreate
from urllib.parse import urlencode
import hashlib


class AsyncPayOK(RequestsClient):
    API_HOST: str = "https://payok.io/"

    def __init__(self, apiKey: str, secretKey: str, apiId: int, shopId: int) -> None:
        """
        Initialize PayOK API client
        :param apiKey: Your api key
        :param secretKey: Your secret key
        :param apiId: Your api id
        :param shopId: Your shop id
        """
        super().__init__()
        self.__apiKey = apiKey
        self.__secretKey = secretKey
        self.__apiId = apiId
        self.__shopId = shopId
        self.__base_url = "https://payok.io/api"
        self.__post_method = "POST"
        self.__payment_name = "payok"
        self.check_values()

    def check_values(self):
        if not self.__secretKey or not self.__apiKey or not self.__apiId or not self.__shopId:
            raise ValueError('No SecretKey, ApiKey, ShopID or ApiID specified')

    async def get_balance(self) -> Balance:
        """Get your balance.
        
        Docs: https://payok.io/cabinet/documentation/doc_api_balance
        """
        url = f"{self.__base_url}/balance"
        
        params = {
            "API_ID": self.__apiId,
            "API_KEY": self.__apiKey,
        }
        response = await self._request(self.__payment_name, self.__post_method, url, data=params)
        
        return Balance(**response)

    async def get_transactions(self, payment: Optional[int] = None, 
                               offset: Optional[int] = None) -> Union[Transaction, List[Transaction]]:
        """Get list of transactions.
        
        Docs: https://payok.io/cabinet/documentation/doc_api_transaction
        
        :param payment: Optional. Payment ID in your system.
        :param offset: Optional. Indentation, skipping the specified number of lines.
        """
        url = f"{self.__base_url}/transaction"
        
        params = {
            "API_ID": self.__apiId,
            "API_KEY": self.__apiKey,
            "shop": self.__shopId,
            "payment": payment,
            "offset": offset,
        }
        self._delete_empty_fields(params)
        response = await self._request(self.__payment_name, self.__post_method, url, data=params)
        if payment:
            return Transaction(**response['1'])
        
        return [Transaction(**transaction) for transaction in response.values()]
        
    async def get_payouts(self, payout_id: Optional[int] = None, offset: Optional[int] = None):
        """Get list of payouts.
        
        Docs: https://payok.io/cabinet/documentation/doc_api_payout
        
        :param payout_id: Payment ID in the Payok system.
        :param offset: Indentation, skipping the specified number of lines.
        """
        url = f"{self.__base_url}/payout"
        params = {
            "API_ID": self.__apiId,
            "API_KEY": self.__apiKey,
            "payout_id": payout_id,
            "offset": offset,
        }
        self._delete_empty_fields(params)
        response = await self._request(self.__payment_name, self.__post_method, url, data=params)
        
        if payout_id:
            return Payout(**response['1'])
        
        return [Payout(**payout) for payout in response.values()]

    async def create_payout(
        self,
        amount: float,
        method: str,
        reciever: str,
        commission_type: str,
        sbp_bank: Optional[str] = None,
        webhook_url: Optional[str] = None
        ) -> CreatePayout:
        """
        Create payout.
        
        Docs: https://payok.io/cabinet/documentation/doc_api_payout_create
        
        :param amount: Payment amount.
        :param method: Special value of the payment method, list of values.
        :param reciever: Details of the payee.
        :param commission_type: The bank for the payment of SBP.
        :param sbp_bank: Type of commission calculation: balance - Commission from the balance sheet, payment - Commission from the payment.
        :param webhook_url: URL for sending a Webhook when the payment status changes.
        """
        url = f"{self.__base_url}/payout_create"
        params = {
            "API_ID": self.__apiId,
            "API_KEY": self.__apiKey,
            "amount": amount,
            "method": method,
            "reciever": reciever,
            "commission_type": commission_type,
            "sbp_bank": sbp_bank,
            "webhook_url": webhook_url
        }
        self._delete_empty_fields(params)
        response = await self._request(self.__payment_name, self.__post_method, url, data=params)
        
        return CreatePayout(remain_balance=response['remain_balance'], payout=PayoutOnCreate(**response['data']))
        
    
    async def create_pay(
        self,
        amount: float,
        payment: Union[int, str],
        currency: Optional[str] = "RUB",
        desc: Optional[str] = 'Description',
        email: Optional[str] = None,
        success_url: Optional[str] = None,
        method: Optional[str] = None,
        lang: Optional[str] = None,
        custom: Optional[str] = None
    ) -> str:
        """Create payform url.
        
        Docs: https://payok.io/cabinet/documentation/doc_payform.php
        
        :param payment: Order number, unique in your system, up to 16 characters. (a-z0-9-_)
        :param amount : Order amount.
        :param currency : ISO 4217 currency. Default is "RUB".
        :param desc : Product name or description.
        :param email : Email Buyer mail. Defaults to None.
        :param success_url: Link to redirect after payment.
        :param method: Payment method
        :param lang: Interface language. RU or EN
        :param custom: Parameter that you want to pass in the notification.
        """

        params = {
            'amount': amount,
            'payment': payment,
            'shop': self.__shopId,
            'currency': currency,
            'desc': desc,
            'email': email,
            'success_url': success_url,
            'method': method,
            'lang': lang,
            'custom': custom
        }

        self._delete_empty_fields(params)

        sign_params = '|'.join(map(
            str,
            [amount, payment, self.__shopId, currency, desc, self.__secretKey]
        )).encode('utf-8')
        sign = hashlib.md5(sign_params).hexdigest()
        params['sign'] = sign

        url = f'{self.API_HOST}/pay?' + urlencode(params)
        return url



from pydantic import BaseModel, Field
from typing import Optional, Union, List


class AccountInfo(BaseModel):
    error: Optional[str] = None
    account: Optional[str] = None
    balance: Optional[Union[int, float]] = None
    currency: Optional[str] = None
    account_type: Optional[str] = None
    identified: Optional[bool] = None
    account_status: Optional[str] = None
    balance_details: Optional[dict] = None
    cards_linked: Optional[list] = None
    
    
class Operation(BaseModel):
    operation_id: Optional[str] = None
    status: Optional[str] = None
    datetime: Optional[str] = None
    title: Optional[str] = None
    pattern_id: Optional[str] = None
    direction: Optional[str] = None
    amount: Optional[Union[int, float]] = None
    label: Optional[str] = None
    type_: Optional[str] = Field(alias="type", default=None)
    amount_due: Optional[Union[int, float]] = None
    fee: Optional[Union[int, float]] = None
    recipient: Optional[str] = None
    recipient_type: Optional[str] = None
    message: Optional[str] = None
    comment: Optional[str] = None
    details: Optional[str] = None
    digital_goods: Optional[Union[str, dict]] = None
    spendingCategories: Optional[list] = []
    amount_currency: Optional[str] = None
    group_id: Optional[str] = None
    categories: Optional[list] = None
    showcase_format: Optional[str] = None
    is_sbp_operation: Optional[bool] = None
    
    
class OperationHistory(BaseModel):
    error: Optional[str] = None
    next_record: Optional[str] = None
    operations: Optional[List[Operation]] = []
    
    
class OperationDetails(BaseModel):
    error: Optional[str] = None
    pattern_id: Optional[str] = None
    group_id: Optional[str] = None
    operation_id: Optional[str] = None
    title: Optional[str] = None
    amount: Optional[Union[int, float]] = None
    direction: Optional[str] = None
    datetime: Optional[str] = None
    status: Optional[str] = None
    type_: Optional[str] = Field(alias="type", default=None)
    available_operations: Optional[list] = []
    amount_due: Optional[Union[int, float]] = None
    fee: Optional[Union[int, float]] = None
    recipient: Optional[str] = None
    recipient_type: Optional[str] = None
    message: Optional[str] = None
    comment: Optional[str] = None
    details: Optional[str] = None
    digital_goods: Optional[str] = None
    spendingCategories: Optional[list] = []
    amount_currency: Optional[str] = None
    categories: Optional[list] = None
    showcase_format: Optional[str] = None
    is_sbp_operation: Optional[bool] = None
    
    
class RequestPayment(BaseModel):
    error: Optional[str] = None
    error_description: Optional[str] = None
    status: Optional[str] = None
    money_source: Optional[dict] = {}
    request_id: Optional[str] = None
    contract_amount: Optional[Union[int, float]] = None
    balance: Optional[Union[int, float]] = None
    recipient_account_status: Optional[str] = None
    recipient_account_type: Optional[str] = None
    account_unblock_uri: Optional[str] = None
    ext_action_uri: Optional[str] = None
    
    
class ProcessPayment(BaseModel):
    error: Optional[str] = None
    error_description: Optional[str] = None
    status: Optional[str] = None
    payment_id: Optional[str] = None
    balance: Optional[Union[int, float]] = None
    invoice_id: Optional[str] = None
    payer: Optional[str] = None
    payee: Optional[str] = None
    credit_amount: Optional[Union[int, float]] = None
    account_unblock_uri: Optional[str] = None
    acs_uri: Optional[str] = None
    acs_params: Optional[dict] = {}
    next_retry: Optional[int] = None
    digital_goods: Optional[Union[str, dict]] = None
    


from pydantic import BaseModel
from typing import Optional, List, Union


class User(BaseModel):
    user_id: Optional[int] = None
    username: Optional[str] = None
    balance: Optional[float] = None
    hold: Optional[float] = None

class Payments(BaseModel):
    payments: Optional[dict] = None
    page: Optional[int] = None
    hasNextPage: Optional[bool] = None


class Invoice(BaseModel):
    amount: Optional[Union[float, int]] = None
    payment_id: Optional[str] = None
    merchant_id: Optional[int] = None
    comment: Optional[str] = None
    additional_data: Optional[str] = None
    url_success: Optional[str] = None
    url_callback: Optional[str] = None
    is_test: Optional[bool] = None
    expires_at: Optional[int] = None
    user_id: Optional[int] = None
    invoice_date: Optional[int] = None
    status: Optional[str] = None
    paid_date: Optional[int] = None
    payer_user_id: Optional[int] = None
    resend_attempts: Optional[int] = None
    invoice_id: Optional[int] = None
    url: Optional[str] = None
    
    
class Invoices(BaseModel):
    invoices: Optional[List[Invoice]] = []
    count: Optional[int] = None
    page: Optional[int] = None
    perPage: Optional[int] = None


from pydantic import BaseModel
from typing import Union, Optional


class Balance(BaseModel):
    balance_rub: Optional[float] = None
    balance_usd: Optional[float] = None


class CreatePayment(BaseModel):
    id: Optional[int] = None
    hash: Optional[str] = None
    url: Optional[str] = None


class Payment(BaseModel):
    id: Optional[int] = None
    order_id: Optional[Union[int, str]] = None
    in_amount: Optional[float] = None
    amount: Optional[float] = None
    currency: Optional[str] = None
    status: Optional[str] = None
    data: Optional[str] = None


class RevokePayment(BaseModel):
    id: Optional[int] = None


class CreateWithdrawRequest(BaseModel):
    id: Optional[int] = None
    status: Optional[str] = None


class CancelWithdrawRequest(BaseModel):
    id: Optional[int] = None
    status: Optional[str] = None


class WithdrawRequest(BaseModel):
    id: Optional[int] = None
    order_id: Optional[Union[int, str]] = None
    amount: Optional[float] = None
    fee: Optional[float] = None
    way: Optional[str] = None
    who_fee: Optional[str] = None
    status: Optional[str] = None



from ..requests import RequestsClient
from typing import Optional, Union
from .models import Order, OrderInfo
import hashlib


class AsyncAPays(RequestsClient):
    API_HOST: str = "https://apays.io/"

    def __init__(self, client_id: int, secret_key: str) -> None:
        """
        Initialize APays API client
        :param client_id: Your APays client ID.
        :param secret_key: Your APays secret key.
        """
        super().__init__()
        self.__client_id = client_id
        self.__secret_key = secret_key
        self.__base_url = "https://apays.io/backend"
        self.__get_method = "GET"
        self.__payment_name = "apays"
        self.check_values()

    def check_values(self):
        if not self.__secret_key or not self.__client_id:
            raise ValueError('No SecretKey or ClientID specified')
    
    def __create_sign(self, order_id: str, amount: Union[float, int] = None) -> str:
        if amount:
            data = f"{order_id}:{amount}:{self.__secret_key}"
        else:
            data = f"{order_id}:{self.__secret_key}"
            
        return hashlib.md5(data.encode()).hexdigest()

    async def create_order(
        self,
        order_id: str,
        amount: Union[int, float],
        email: Optional[str] = None,
    ) -> Order:
        """Creates a payment and returns a link for the buyer to pay.
        
        :param order_id: Unique order ID.
        :param amount: Amount in kopecks.
        :param email: Optional. Buyer's email address. If set, it will not be requested.
        
        Docs: https://docs.apays.io/lets-start/api/create-payment
        """
        params = {
            "client_id": self.__client_id,
            "order_id": order_id,
            "amount": amount,
            "sign": self.__create_sign(order_id, amount),
            "email": email,
        }
        self._delete_empty_fields(params)
        url = f"{self.__base_url}/create_order"

        response = await self._request(
            self.__payment_name,
            self.__get_method,
            url,
            params=params,
        )
        return Order(**response)
    
    async def get_order(self, order_id: str) -> OrderInfo:
        """Receiving payment.

        :param order_id: Order ID.
        
        Docs: https://docs.apays.io/lets-start/api/check-status-payment
        """
        params = {
            "client_id": self.__client_id,
            "order_id": order_id,
            "sign": self.__create_sign(order_id)
        }
        url = f"{self.__base_url}/get_order"

        response = await self._request(
            self.__payment_name,
            self.__get_method,
            url,
            params=params,
        )
        return OrderInfo(**response)